//================================================================================================
// $Id: cginstrument.h 6464 2023-02-27 20:03:38Z aron $
//
// Project:
//		Welltype Compiler
// Contents:
//		Code generator instrument
// Authors:
//		Aron Barath, 2014-2023
//================================================================================================
// Copyright (c) Aron Barath
//================================================================================================

#ifndef WTC_CGINSTRUMENT_H
#define WTC_CGINSTRUMENT_H

#include <string>
#include <map>
#include <set>
#include <vector>
#include <deque>
#include "../wtparser/uniqueid.h"
#include "../wtruntime/wtruntime.h"

namespace wtast
{
class FunctionSignature;
} // wtast

namespace wtc
{
class ProblemHandler;
} // wtc

namespace wtc
{
namespace codegen
{

class CodeSection;
class DLabelSection;

struct Instrument
{
	enum Kind
	{
		IK_SINGLE, // this is a single precompiled intruction (opcode+imm8)
		IK_DOUBLE, // this is a double precompiled intruction (opcode+imm8+imm16)
		IK_GLOBAL, // this instrument does not contain opcode ('data' is a global label)
		IK_LABEL,  // this instrument does not contain opcode ('data' is a label)
		IK_JMPI,   // 'data' refers to an instrument (jmp.i only)
		IK_JMPL,   // 'data' refers to an instrument (jmp.l only)
		IK_JTI,    // 'data' refers to an instrument (jump when true; jt.i only)
		IK_JTL,    // 'data' refers to an instrument (jump when true; jt.l only)
		IK_JFI,    // 'data' refers to an instrument (jump when false; jf.i only)
		IK_JFL,    // 'data' refers to an instrument (jump when false; jf.l only)
		IK_ENTER,  // 'data' refers to a function name to ENTER (eg. full name)
		IK_MKFP,   // 'data' refers to a function name to MKFP (eg. full name)
		IK_EXHAN,  // 'data' refers to an instrument
	};

	Kind        kind;
	wt_u8       opcode;
	wt_u8       imm8;
	wt_u16      imm16;
	std::string data;

	Instrument() { }
	Instrument(const Instrument & rhs) : kind(rhs.kind), opcode(rhs.opcode), imm8(rhs.imm8), imm16(rhs.imm16), data(rhs.data) { }
	Instrument(Kind k, wt_u8 o) : kind(k), opcode(o), imm8(0), imm16(0) { }
	Instrument(Kind k, wt_u8 o, wt_u8 i8) : kind(k), opcode(o), imm8(i8), imm16(0) { }
	Instrument(Kind k, wt_u8 o, wt_u8 i8, wt_u16 i16) : kind(k), opcode(o), imm8(i8), imm16(i16) { }
	Instrument(Kind k, const std::string & d) : kind(k), opcode(0), imm8(0), imm16(0), data(d) { }
	Instrument(Kind k, wt_u8 o, const std::string & d) : kind(k), opcode(o), imm8(0), imm16(0), data(d) { }
};

class InstrumentProblemHandler
{
public:
	InstrumentProblemHandler() { }
	virtual ~InstrumentProblemHandler() { }
	virtual void InstrumentNote(const std::string & msg) = 0;
	virtual void InstrumentWarning(const std::string & msg) = 0;
	virtual void InstrumentError(const std::string & msg) = 0;
	virtual wtc::ProblemHandler & GetProblemHandler() = 0;
};

class InstrumentEngine
{
	typedef std::deque<Instrument> instruments_t;
	typedef std::map<wtparser::UniqueId, instruments_t> map_uniqueid_ins_t;
	typedef std::map<wtparser::UniqueId, std::string> map_uniqueid_str_t;
	typedef std::set<wtparser::UniqueId> set_uniqueid_t;
	typedef std::map<std::string, wt_u32> map_label_pc_t;
	typedef std::map<wt_u32, std::string> map_pc_resolve_t;
	typedef InstrumentProblemHandler probhan_t;

	instruments_t      mInstruments;
	map_uniqueid_ins_t mEXEC_Table;
	map_uniqueid_str_t mBuiltinNames;
	std::string        mLastGlobal;
	unsigned int       mNextLabelId;
	map_label_pc_t     mMapLabelToPC;
	map_label_pc_t     mMapCanonToPC;
	probhan_t*         mProblemHandler;
	set_uniqueid_t     mInstantiateBuiltins;

	void Resolve(CodeSection* cs, map_pc_resolve_t & resolve, wt_u8 opcode, bool relative, bool single);
	void Make_Generic(wt_u32 addr, wt_u8 op_i, wt_u8 op_l);
	void Make_Generic(wt_u32 addr, wt_u8 op_i, wt_u8 op_l, wt_u8 op_s);
	std::string TruncateCanonical(const std::string & name);

public:
	InstrumentEngine(InstrumentProblemHandler* probhan);
	~InstrumentEngine() { }

	std::string GenLabel();

	std::size_t GetInstrumentCount() const { return mInstruments.size(); }

	void Add(const Instrument & ins) { mInstruments.push_back(ins); }
	void AddSingle(wt_u8 opcode, wt_u8 imm8=0) { mInstruments.push_back(Instrument(Instrument::IK_SINGLE, opcode, imm8)); }
	void AddDouble(wt_u8 opcode, wt_u8 imm8, wt_u16 imm16) { mInstruments.push_back(Instrument(Instrument::IK_DOUBLE, opcode, imm8, imm16)); }
	void AddGlobal(const std::string & label) { mInstruments.push_back(Instrument(Instrument::IK_GLOBAL, label)); mLastGlobal = label; mNextLabelId = 1; }
	void AddLabel(const std::string & label) { mInstruments.push_back(Instrument(Instrument::IK_LABEL, label)); }
	void AddJmpI(const std::string & label) { mInstruments.push_back(Instrument(Instrument::IK_JMPI, label)); }
	void AddJmpL(const std::string & label) { mInstruments.push_back(Instrument(Instrument::IK_JMPL, label)); }
	void AddJTI(const std::string & label) { mInstruments.push_back(Instrument(Instrument::IK_JTI, label)); }
	void AddJTL(const std::string & label) { mInstruments.push_back(Instrument(Instrument::IK_JTL, label)); }
	void AddJFI(const std::string & label) { mInstruments.push_back(Instrument(Instrument::IK_JFI, label)); }
	void AddJFL(const std::string & label) { mInstruments.push_back(Instrument(Instrument::IK_JFL, label)); }
	void AddEnter(const std::string & label) { mInstruments.push_back(Instrument(Instrument::IK_ENTER, label)); }
	void AddMkfp(const std::string & label) { mInstruments.push_back(Instrument(Instrument::IK_MKFP, label)); }
	void AddExhan(const std::string & label) { mInstruments.push_back(Instrument(Instrument::IK_EXHAN, label)); }

	void InsertSingle(std::size_t idx, wt_u8 opcode, wt_u8 imm8=0) { mInstruments.insert(mInstruments.begin()+idx, Instrument(Instrument::IK_SINGLE, opcode, imm8)); }
	void InsertDouble(std::size_t idx, wt_u8 opcode, wt_u8 imm8, wt_u16 imm16) { mInstruments.insert(mInstruments.begin()+idx, Instrument(Instrument::IK_DOUBLE, opcode, imm8, imm16)); }

	void AddBuiltinInstantiation(wtparser::UniqueId unique_id, const std::string & name) { mInstantiateBuiltins.insert(unique_id); mBuiltinNames[unique_id] = name; }

	bool ReplaceLast(const wt_u8 opcode_to_replace, wt_u8 new_opcode);
	bool LastIsWith(const wt_u8 opcode) const;
	Instrument GetLastInstrument() const { return mInstruments.back(); }
	Instrument GetInstrument(std::size_t idx) const { return mInstruments[idx]; }
	void RemoveInstrument(std::size_t idx) { mInstruments.erase(mInstruments.begin()+idx); }
	void RemoveLastInstrument() { mInstruments.pop_back(); }

	void MakePushImm(wt_u32 value);
	void MakePushImm(wt_u64 value);
	void MakeST(wt_u32 size);
	void MakeLD(wt_u32 size);
	void MakePBA(wt_u32 addr)   { Make_Generic(addr, WT_TOMB_OP_PBA_I, WT_TOMB_OP_PBA_L, WT_TOMB_OP_PBA_S); }
	void MakeDVTA(wt_u32 addr)  { Make_Generic(addr, WT_TOMB_OP_DVTA_I, WT_TOMB_OP_DVTA_L, WT_TOMB_OP_DVTA_S); }
	void MakePVT(wt_u32 addr)   { Make_Generic(addr, WT_TOMB_OP_PVT_I, WT_TOMB_OP_PVT_L, WT_TOMB_OP_PVT_S); }
	void MakePDSP(wt_u32 offs)  { Make_Generic(offs, WT_TOMB_OP_PDSP_I, WT_TOMB_OP_PDSP_L); }
	void MakeDUP(wt_u32 size)   { Make_Generic(size, WT_TOMB_OP_DUP_I, WT_TOMB_OP_DUP_L); }
	void MakePOP(wt_u32 size)   { Make_Generic(size, WT_TOMB_OP_POP_I, WT_TOMB_OP_POP_L); }
	void MakeDLE(wt_u32 offs)   { Make_Generic(offs, WT_TOMB_OP_DLE_I, WT_TOMB_OP_DLE_L, WT_TOMB_OP_DLE_S); }
	void MakeRECN(wt_u32 size)  { Make_Generic(size, WT_TOMB_OP_RECN_I, WT_TOMB_OP_RECN_L, WT_TOMB_OP_RECN_S); }
	void MakeRECFR(wt_u32 offs) { Make_Generic(offs, WT_TOMB_OP_RECFR_I, WT_TOMB_OP_RECFR_L, WT_TOMB_OP_RECFR_S); }
	void MakeSEQN(wt_u32 size)  { Make_Generic(size, WT_TOMB_OP_SEQN_I, WT_TOMB_OP_SEQN_L, WT_TOMB_OP_SEQN_S); }
	void MakeSEQIR(wt_u32 indx) { Make_Generic(indx, WT_TOMB_OP_SEQIR_I, WT_TOMB_OP_SEQIR_L, WT_TOMB_OP_SEQIR_S); }
	void MakeSTRCR(wt_u32 indx) { Make_Generic(indx, WT_TOMB_OP_STRCR_I, WT_TOMB_OP_STRCR_L, WT_TOMB_OP_STRCR_S); }
	void MakeEXEC(wt_u32 indx)  { Make_Generic(indx, WT_TOMB_OP_EXEC_I, WT_TOMB_OP_EXEC_L, WT_TOMB_OP_EXEC_S); }
	void MakeEXEC(wtast::FunctionSignature* sig);
	void MakeENTER(wtast::FunctionSignature* sig);

	CodeSection* Compile(const std::string & cs_name, wt_u32 cs_addr, DLabelSection* dlabs, bool all_labels);
};

} // codegen
} // wtc

#endif // WTC_CGINSTRUMENT_H
