//================================================================================================
// $Id: codegen_io.cpp 6466 2023-02-27 20:09:00Z aron $
//
// Project:
//		Welltype Compiler
// Contents:
//		Code generator
// Authors:
//		Aron Barath, 2015-2023
//================================================================================================
// Copyright (c) Aron Barath
//================================================================================================

#include "codegen.h"
#include "tinyframe.h"
#include "wtruntime/wtruntime.h"
#include "wtruntime/types.h"
#include "wtruntime/stdlib.h"
#include "wtparser/astvisitor.h"
#include "wtparser/ast/ast_type.h"
#include "wtparser/ast/ast_typeinfo.h"
#include "wtparser/ast/ast_seqtype.h"
#include "wtparser/ast/ast_functionsignature.h"
#include "wtparser/ast/ast_functionargument.h"

namespace wtc
{
namespace codegen
{

namespace
{
namespace layout
{

static const wt_u32 counter_size = sizeof(wt_u32);
static const wt_u32 counter_offs = 0;
static const wt_u32 handle_size = sizeof(wt_handle);
static const wt_u32 handle_offs = counter_size;

static const wt_u32 record_size = counter_size + handle_size;

} // layout
} // <anon>

void CodeGen::GenIO_FileDtor()
{
	GenIO__GenericDtor(WTMP_FILECLOSE);
}

void CodeGen::GenIO_FileCopy()
{
	GenIO__GenericCopy();
}

void CodeGen::GenIO_DirDtor()
{
	GenIO__GenericDtor(WTMP_DIRCLOSE);
}

void CodeGen::GenIO_DirCopy()
{
	GenIO__GenericCopy();
}

void CodeGen::GenIO__GenericDtor(wt_u32 exec_code)
{
	const std::string del_label  = mEngine.GenLabel();
	const std::string keep_label = mEngine.GenLabel();

	// decrement the refcounter

	// [Addr]
	mEngine.MakeDUP(sizeof(wt_vmptr));
	// [Addr, Addr]
	mEngine.MakeRECFR(layout::counter_offs);
	// [Addr, &RefCount]
	mEngine.MakeLD(sizeof(wt_u32));
	// [Addr, RefCount]
	mEngine.AddSingle(WT_TOMB_OP_SUB_I, 1);
	// [Addr, NewRefCount]
	mEngine.MakePDSP(sizeof(wt_u32));
	// [Addr, NewRefCount, &Addr]
	mEngine.MakeLD(sizeof(wt_vmptr));
	// [Addr, NewRefCount, Addr]
	mEngine.MakeRECFR(layout::counter_offs);
	// [Addr, NewRefCount, &RefCount]
	mEngine.MakeST(sizeof(wt_u32));
	// [Addr]

	// load refcounter

	// [Addr]
	mEngine.MakeDUP(sizeof(wt_vmptr));
	// [Addr, Addr]
	mEngine.MakeRECFR(layout::counter_offs);
	// [Addr, &RefCount]
	mEngine.MakeLD(sizeof(wt_u32));
	// [Addr, RefCount]
	mEngine.AddJFI(del_label);
	// [Addr]
	mEngine.AddSingle(WT_TOMB_OP_POP_I, sizeof(wt_vmptr));
	// []
	mEngine.AddJmpL(keep_label);

	// really delete it

	mEngine.AddLabel(del_label);
	// [Addr]
	mEngine.MakeDUP(sizeof(wt_vmptr));
	// [Addr, Addr]
	mEngine.MakeRECFR(layout::handle_offs);
	// [Addr, &Handle]
	mEngine.MakeLD(layout::handle_size);
	// [Addr, Handle]
	mEngine.MakeEXEC(exec_code);
	// [Addr]
	mEngine.AddSingle(WT_TOMB_OP_RECD);
	// []
	mEngine.AddLabel(keep_label);
}

void CodeGen::GenIO__GenericCopy()
{
	//mEngine.AddSingle(WT_TOMB_OP_LOCK); // TODO
	// [Addr]
	mEngine.MakeDUP(sizeof(wt_vmptr));
	// [Addr, Addr]
	mEngine.MakeRECFR(layout::counter_offs);
	// [Addr, &RefCount]
	mEngine.MakeLD(sizeof(wt_u32));
	// [Addr, RefCount]
	mEngine.AddSingle(WT_TOMB_OP_ADD_I, 1);
	// [Addr, NewRefCount]
	mEngine.MakePDSP(sizeof(wt_u32));
	// [Addr, NewRefCount, &Addr]
	mEngine.MakeLD(sizeof(wt_vmptr));
	// [Addr, NewRefCount, Addr]
	mEngine.MakeRECFR(layout::counter_offs);
	// [Addr, NewRefCount, &RefCount]
	mEngine.MakeST(sizeof(wt_u32));
	// [Addr]
	//mEngine.AddSingle(WT_TOMB_OP_ULOCK); // TODO
}

void CodeGen::GenIO_fstat(wtast::FunctionSignature* sig)
{
	mEngine.AddGlobal(sig->GetCanonicalName(false, true));
	mEngine.MakeRECFR(layout::handle_offs);
	mEngine.MakeLD(layout::handle_size);
	mEngine.MakeEXEC(WTMP_FILESTATF);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenIO_fopen(wtast::FunctionSignature* sig)
{
	TinyFrame tf;

	const wt_u32 rec_offs = tf.Add(sizeof(wt_vmptr));

	mEngine.AddGlobal(sig->GetCanonicalName(false, true));
	mEngine.MakeDVTA(tf.Size());
	mEngine.AddSingle(WT_TOMB_OP_RECN_I, layout::record_size);
	mEngine.MakePVT(rec_offs);
	mEngine.MakeST(sizeof(wt_vmptr));

	mEngine.MakeEXEC(WTMP_FILEOPEN);
	mEngine.MakePVT(rec_offs);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.MakeRECFR(layout::handle_offs);
	mEngine.MakeST(layout::handle_size);

	mEngine.AddSingle(WT_TOMB_OP_PWZ_I, 1);
	mEngine.MakePVT(rec_offs);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.MakeRECFR(layout::counter_offs);
	mEngine.MakeST(layout::counter_size);

	mEngine.MakePVT(rec_offs);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
	mEngine.AddMkfp(GetDestructorName(sig->GetRet(0)));
	mEngine.AddSingle(WT_TOMB_OP_TPU);
	mEngine.AddSingle(WT_TOMB_OP_DVTF);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenIO_feof(wtast::FunctionSignature* sig)
{
	mEngine.AddGlobal(sig->GetCanonicalName(false, true));
	mEngine.MakeRECFR(layout::handle_offs);
	mEngine.MakeLD(layout::handle_size);
	mEngine.MakeEXEC(WTMP_FILEEOF);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenIO_freadstr(wtast::FunctionSignature* sig)
{
	mEngine.AddGlobal(sig->GetCanonicalName(false, true));

	if(1==sig->GetArgc())
	{
		mEngine.MakeRECFR(layout::handle_offs);
		mEngine.MakeLD(layout::handle_size);
		mEngine.AddSingle(WT_TOMB_OP_PWZ_I, '\0');
		mEngine.MakeEXEC(WTMP_FILEREADSTR);
		mEngine.AddSingle(WT_TOMB_OP_LEAVE);
	}
	else
	{
		mEngine.MakeDVTA(sizeof(wt_char));
		mEngine.MakePVT(0);
		mEngine.MakeST(sizeof(wt_char));
		mEngine.MakeRECFR(layout::handle_offs);
		mEngine.MakeLD(layout::handle_size);
		mEngine.MakePVT(0);
		mEngine.MakeLD(sizeof(wt_char));
		mEngine.MakeEXEC(WTMP_FILEREADSTR);
		mEngine.AddSingle(WT_TOMB_OP_DVTF);
		mEngine.AddSingle(WT_TOMB_OP_LEAVE);
	}
}

void CodeGen::GenIO_freadln(wtast::FunctionSignature* sig)
{
	mEngine.AddGlobal(sig->GetCanonicalName(false, true));
	mEngine.MakeRECFR(layout::handle_offs);
	mEngine.MakeLD(layout::handle_size);
	mEngine.AddSingle(WT_TOMB_OP_PWZ_I, '\n');
	mEngine.MakeEXEC(WTMP_FILEREADSTR);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenIO__GenericFread(wtast::FunctionSignature* sig)
{
	wtast::Type* type      = sig->GetRet(0);
	const wt_u32 load_size = type->GetLoadSize();
	const wt_u32 io_size   = (WTT_CHAR==type->GetTypeId() || WTT_BOOL==type->GetTypeId()) ? 1 : load_size;

	mEngine.AddGlobal(sig->GetCanonicalName(false, true));
	mEngine.MakeDVTA(std::max(wt_u32(4), load_size));

	mEngine.MakeRECFR(layout::handle_offs);
	mEngine.MakeLD(layout::handle_size);
	mEngine.MakePVT(0);
	mEngine.AddSingle(WT_TOMB_OP_PWZ_I, io_size);
	mEngine.MakeEXEC(WTMP_FILEREAD);
	mEngine.MakePVT(0);
	mEngine.MakeLD(load_size);

	if(4>io_size)
	{
		// back to native order
		mEngine.MakeEXEC(WTMP_LSWAP32);
	}

	mEngine.AddSingle(WT_TOMB_OP_DVTF);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenIO_fread(wtast::FunctionSignature* sig)
{
	wtast::Type* type = sig->GetArg(1)->GetType();
	const std::size_t elem_size = static_cast<wtast::SeqType*>(type)->GetElementType()->GetLoadSize();

	mEngine.AddGlobal(sig->GetCanonicalName(false, true));

	mEngine.MakeDVTA(sizeof(wt_vmptr));
	mEngine.MakePVT(0);
	mEngine.MakeST(sizeof(wt_vmptr));

	// 1st: file handle (file record is still on the top of the stack)
	mEngine.MakeRECFR(layout::handle_offs);
	mEngine.MakeLD(layout::handle_size);

	// 2nd: array start address
	mEngine.MakePVT(0);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_SEQIR_I, 0);

	// 3rd: total size
	mEngine.MakePVT(0);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_SEQL);
	switch(elem_size)
	{
	case 1: break;
	case 2: mEngine.AddSingle(WT_TOMB_OP_SHL_I, 1); break;
	case 4: mEngine.AddSingle(WT_TOMB_OP_SHL_I, 2); break;
	case 8: mEngine.AddSingle(WT_TOMB_OP_SHL_I, 3); break;
	default: break;
	}

	// call filereadr
	mEngine.MakeEXEC(WTMP_FILEREADR);

	switch(elem_size)
	{
	case 1: break;
	case 2: mEngine.AddSingle(WT_TOMB_OP_SHR_I, 1); break;
	case 4: mEngine.AddSingle(WT_TOMB_OP_SHR_I, 2); break;
	case 8: mEngine.AddSingle(WT_TOMB_OP_SHR_I, 3); break;
	default: break;
	}

	// cleanup
	mEngine.AddSingle(WT_TOMB_OP_DVTF);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenIO_freadchar(wtast::FunctionSignature* sig)
{
	GenIO__GenericFread(sig);
}

void CodeGen::GenIO_freadbool(wtast::FunctionSignature* sig)
{
	GenIO__GenericFread(sig);
}

void CodeGen::GenIO_freadi8(wtast::FunctionSignature* sig)
{
	GenIO__GenericFread(sig);
}

void CodeGen::GenIO_freadu8(wtast::FunctionSignature* sig)
{
	GenIO__GenericFread(sig);
}

void CodeGen::GenIO_freadi16(wtast::FunctionSignature* sig)
{
	GenIO__GenericFread(sig);
}

void CodeGen::GenIO_freadu16(wtast::FunctionSignature* sig)
{
	GenIO__GenericFread(sig);
}

void CodeGen::GenIO_freadi32(wtast::FunctionSignature* sig)
{
	GenIO__GenericFread(sig);
}

void CodeGen::GenIO_freadu32(wtast::FunctionSignature* sig)
{
	GenIO__GenericFread(sig);
}

void CodeGen::GenIO_freadf32(wtast::FunctionSignature* sig)
{
	GenIO__GenericFread(sig);
}

void CodeGen::GenIO_freadi64(wtast::FunctionSignature* sig)
{
	GenIO__GenericFread(sig);
}

void CodeGen::GenIO_freadu64(wtast::FunctionSignature* sig)
{
	GenIO__GenericFread(sig);
}

void CodeGen::GenIO_freadf64(wtast::FunctionSignature* sig)
{
	GenIO__GenericFread(sig);
}

void CodeGen::GenIO_fwrite(wtast::FunctionSignature* sig)
{
	wtast::Type* type = sig->GetArg(1)->GetType();

	if(!type->IsSequence())
	{
		const wt_u32 load_size = type->GetLoadSize();
		const wt_u32 io_size   = (WTT_CHAR==type->GetTypeId() || WTT_BOOL==type->GetTypeId()) ? 1 : load_size;

		if(WTT_STRING==type->GetTypeId())
		{
			GenIO__GenericFwriteString(sig, '\0');
			return;
		}

		mEngine.AddGlobal(sig->GetCanonicalName(false, true));
		mEngine.AddSingle(WT_TOMB_OP_DVTA_I, std::max(wt_u32(4), load_size));

		if(4>io_size)
		{
			// back to native order
			mEngine.MakeEXEC(WTMP_LSWAP32);
		}
		mEngine.MakePVT(0);

		mEngine.MakeST(load_size);
		mEngine.MakeRECFR(layout::handle_offs);
		mEngine.MakeLD(layout::handle_size);
		mEngine.MakePVT(0);
		mEngine.AddSingle(WT_TOMB_OP_PWZ_I, io_size);

		mEngine.MakeEXEC(WTMP_FILEWRITE);

		mEngine.AddSingle(WT_TOMB_OP_DVTF);
		mEngine.AddSingle(WT_TOMB_OP_LEAVE);
	}
	else
	{
		const std::size_t elem_size = static_cast<wtast::SeqType*>(type)->GetElementType()->GetLoadSize();

		mEngine.AddGlobal(sig->GetCanonicalName(false, true));

		mEngine.MakeDVTA(sizeof(wt_vmptr));
		mEngine.MakePVT(0);
		mEngine.MakeST(sizeof(wt_vmptr));

		// 1st: file handle (file record is still on the top of the stack)
		mEngine.MakeRECFR(layout::handle_offs);
		mEngine.MakeLD(layout::handle_size);

		// 2nd: array start address
		mEngine.MakePVT(0);
		mEngine.MakeLD(sizeof(wt_vmptr));
		mEngine.AddSingle(WT_TOMB_OP_SEQIR_I, 0);

		// 3rd: total size
		mEngine.MakePVT(0);
		mEngine.MakeLD(sizeof(wt_vmptr));
		mEngine.AddSingle(WT_TOMB_OP_SEQL, 0);
		switch(elem_size)
		{
		case 1: break;
		case 2: mEngine.AddSingle(WT_TOMB_OP_SHL_I, 1); break;
		case 4: mEngine.AddSingle(WT_TOMB_OP_SHL_I, 2); break;
		case 8: mEngine.AddSingle(WT_TOMB_OP_SHL_I, 3); break;
		default: break;
		}

		// call filewrite
		mEngine.MakeEXEC(WTMP_FILEWRITE);

		// cleanup
		mEngine.AddSingle(WT_TOMB_OP_DVTF);
		mEngine.AddSingle(WT_TOMB_OP_LEAVE);
	}
}

void CodeGen::GenIO_fwriteln(wtast::FunctionSignature* sig)
{
	GenIO__GenericFwriteString(sig, '\n');
}

void CodeGen::GenIO__GenericFwriteString(wtast::FunctionSignature* sig, wt_char ch)
{
	const wt_u32 load_size = sig->GetArg(1)->GetType()->GetLoadSize();

	mEngine.AddGlobal(sig->GetCanonicalName(false, true));
	mEngine.AddSingle(WT_TOMB_OP_DVTA_I, load_size);
	mEngine.MakePVT(0);
	mEngine.MakeST(load_size);
	mEngine.MakeRECFR(layout::handle_offs);
	mEngine.MakeLD(layout::handle_size);
	mEngine.MakePVT(0);
	mEngine.MakeLD(load_size);
	mEngine.AddSingle(WT_TOMB_OP_PWZ_I, (wt_u8)ch);
	mEngine.MakeEXEC(WTMP_FILEWRITESTR);
	mEngine.AddSingle(WT_TOMB_OP_DVTF);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenIO_opendir(wtast::FunctionSignature* sig)
{
	const wt_u32 rec_offs = 0;
	const wt_u32 dvt_size = sizeof(wt_vmptr);

	mEngine.AddGlobal(sig->GetCanonicalName(false, true));
	mEngine.AddSingle(WT_TOMB_OP_DVTA_I, dvt_size);
	mEngine.AddSingle(WT_TOMB_OP_RECN_I, layout::record_size);
	mEngine.MakePVT(rec_offs);
	mEngine.MakeST(sizeof(wt_vmptr));

	mEngine.MakeEXEC(WTMP_DIROPEN);
	mEngine.MakePVT(rec_offs);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.MakeRECFR(layout::handle_offs);
	mEngine.MakeST(layout::handle_size);

	mEngine.AddSingle(WT_TOMB_OP_PWZ_I, 1);
	mEngine.MakePVT(rec_offs);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.MakeRECFR(layout::counter_offs);
	mEngine.MakeST(layout::counter_size);

	mEngine.MakePVT(rec_offs);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
	mEngine.AddMkfp(GetDestructorName(sig->GetRet(0)));
	mEngine.AddSingle(WT_TOMB_OP_TPU);
	mEngine.AddSingle(WT_TOMB_OP_DVTF);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenIO_readdir(wtast::FunctionSignature* sig)
{
	mEngine.AddGlobal(sig->GetCanonicalName(false, true));
	mEngine.MakeRECFR(layout::handle_offs);
	mEngine.MakeLD(layout::handle_size);
	mEngine.MakeEXEC(WTMP_DIRREAD);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenIO_dirgetname(wtast::FunctionSignature* sig)
{
	mEngine.AddGlobal(sig->GetCanonicalName(false, true));
	mEngine.MakeRECFR(layout::handle_offs);
	mEngine.MakeLD(layout::handle_size);
	mEngine.MakeEXEC(WTMP_DIRGETNAME);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenIO_dirgettype(wtast::FunctionSignature* sig)
{
	mEngine.AddGlobal(sig->GetCanonicalName(false, true));
	mEngine.MakeRECFR(layout::handle_offs);
	mEngine.MakeLD(layout::handle_size);
	mEngine.MakeEXEC(WTMP_DIRGETTYPE);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

} // codegen
} // wtc
