#include "config_private.h"
#include "microprog.h"
#include "wtruntime.h"
#include "version.h"
#include "profile.h"
#include "program.h"
#include "ctx.h"
#include "processor.h"
#include "objects.h"
#include "stdlib.h"
#include "convert.h"
#include "byteorder.h"
#include "unicode.h"
#include "iocall.h"
#include <string.h>
#include <ctype.h>
#if WTRUNTIME_ENABLE_FLOAT32 || WTRUNTIME_ENABLE_FLOAT64
#  include <math.h>
#  include <float.h>
#endif /* WTRUNTIME_ENABLE_FLOAT32 || WTRUNTIME_ENABLE_FLOAT64 */

#define __WTR_LIST_MICROPROGRAM(lowercase,UPPERCASE) \
	__WTR_NAME_MICROPROGRAM(lowercase,UPPERCASE),

#if !WTRUNTIME_ENABLE_FLOAT32 || !WTRUNTIME_ENABLE_FLOAT64
#  define wtr_mp_cf32f64        wtr_mp_undefined
#  define wtr_mp_cf64f32        wtr_mp_undefined
#endif /* !WTRUNTIME_ENABLE_FLOAT32 || !WTRUNTIME_ENABLE_FLOAT64 */
#if !WTRUNTIME_ENABLE_FLOAT32
#  define wtr_mp_randf          wtr_mp_undefined
#  define wtr_mp_f32sqrt        wtr_mp_undefined
#  define wtr_mp_f32log         wtr_mp_undefined
#  define wtr_mp_f32exp         wtr_mp_undefined
#  define wtr_mp_f32sin         wtr_mp_undefined
#  define wtr_mp_f32cos         wtr_mp_undefined
#  define wtr_mp_f32tan         wtr_mp_undefined
#  define wtr_mp_f32asin        wtr_mp_undefined
#  define wtr_mp_f32acos        wtr_mp_undefined
#  define wtr_mp_f32atan        wtr_mp_undefined
#  define wtr_mp_f32atan2       wtr_mp_undefined
#  define wtr_mp_f32pow         wtr_mp_undefined
#  define wtr_mp_cf32string     wtr_mp_undefined
#  define wtr_mp_cstringf32     wtr_mp_undefined
#  define wtr_mp_cu32f32        wtr_mp_undefined
#  define wtr_mp_ci32f32        wtr_mp_undefined
#  define wtr_mp_cf32i32        wtr_mp_undefined
#  define wtr_mp_cf32u32        wtr_mp_undefined
#  define wtr_mp_ci64f32        wtr_mp_undefined
#  define wtr_mp_cu64f32        wtr_mp_undefined
#  define wtr_mp_cf32i64        wtr_mp_undefined
#  define wtr_mp_cf32u64        wtr_mp_undefined
#endif /* !WTRUNTIME_ENABLE_FLOAT32 */
#if !WTRUNTIME_ENABLE_FLOAT64
#  define wtr_mp_f64sqrt        wtr_mp_undefined
#  define wtr_mp_f64log         wtr_mp_undefined
#  define wtr_mp_f64exp         wtr_mp_undefined
#  define wtr_mp_f64sin         wtr_mp_undefined
#  define wtr_mp_f64cos         wtr_mp_undefined
#  define wtr_mp_f64tan         wtr_mp_undefined
#  define wtr_mp_f64asin        wtr_mp_undefined
#  define wtr_mp_f64acos        wtr_mp_undefined
#  define wtr_mp_f64atan        wtr_mp_undefined
#  define wtr_mp_f64atan2       wtr_mp_undefined
#  define wtr_mp_f64pow         wtr_mp_undefined
#  define wtr_mp_cf64string     wtr_mp_undefined
#  define wtr_mp_cstringf64     wtr_mp_undefined
#  define wtr_mp_cu32f64        wtr_mp_undefined
#  define wtr_mp_ci32f64        wtr_mp_undefined
#  define wtr_mp_cf64i32        wtr_mp_undefined
#  define wtr_mp_cf64u32        wtr_mp_undefined
#  define wtr_mp_ci64f64        wtr_mp_undefined
#  define wtr_mp_cu64f64        wtr_mp_undefined
#  define wtr_mp_cf64i64        wtr_mp_undefined
#  define wtr_mp_cf64u64        wtr_mp_undefined
#endif /* !WTRUNTIME_ENABLE_FLOAT64 */

WTR_PFN_MICROPROGRAM wtr_microprogtab[] =
{
	WTR_MICROPROGRAMS(__WTR_LIST_MICROPROGRAM)
};

#define CHECK_POP(_ctx_,_size_) \
	do \
	{ \
		if(WTR_CTX_DS_SIZE(_ctx_)>=(_size_)) \
		{ \
			/* branch prediction */ \
		} \
		else \
		{ \
			ctx->last_exception = StackOverflowException; \
			return WTRR_RAISE; \
		} \
	} \
	while(0)

#define CHECK_PUSH(_ctx_,_size_) \
	do \
	{ \
		if(WTR_CTX_DS_SPACE(_ctx_)>=(_size_)) \
		{ \
			/* branch prediction */ \
		} \
		else \
		{ \
			ctx->last_exception = StackOverflowException; \
			return WTRR_RAISE; \
		} \
	} \
	while(0)

#define STR_REGISTER(_prof_,_prog_,_ctx_,_str_,_vmaddr_) \
	do \
	{ \
		struct wtr_objentry ent; \
		\
		ent.obj     = (_vmaddr_); \
		ent.pfn.ptr = (_prog_); \
		ent.pfn.ext = (_prof_)->pc_strd_leave; \
		\
		if(!wtr_dynstack_push(&(_ctx_)->tos, &ent)) \
		{ \
			wtr_str_delete((_prog_), (_str_)); \
			(_ctx_)->last_exception = OutOfMemoryException; \
			return WTRR_RAISE; \
		} \
	} \
	while(0)

static const wt_u8 _squares_u8[] =
{
	  0,   1,   4,   9,
	 16,  25,  36,  49,
	 64,  81, 100, 121,
	144, 169, 196, 225,
};

static wt_u8 sqrt_u8(wt_u8 x)
{
	const wt_u8* p = _squares_u8;

	if(p[8] <= x) { p += 8; }
	if(p[4] <= x) { p += 4; }
	if(p[2] <= x) { p += 2; }
	if(p[1] <= x) { p += 1; }

	return (wt_u8)(p - _squares_u8);
}

static wt_i8 sqrt_i8(wt_i8 n)
{
	if(0>=n)
	{
		return 0;
	}

	return sqrt_u8(n);
}

static wt_u32 sqrt_u32(wt_u32 remainder)
{
	wt_u32 place = ((wt_u32)1) << (sizeof(wt_u32) * 8 - 2); /* calculated by compiler = same runtime as: place = 0x40000000 */
	wt_u32 root  = 0;

	while(place>remainder)
	{
		place >>= 2;
	}

	while(place)
	{
		if(remainder>=root+place)
		{
			remainder -= root+place;
			root += place << 1;
		}

		root >>= 1;
		place >>= 2;
	}

	return root;
}

static wt_i32 sqrt_i32(wt_i32 n)
{
	if(0>=n)
	{
		return 0;
	}

	return sqrt_u32((wt_u32)n);
}

static wt_i16 sqrt_i16(wt_i16 n)
{
	if(0>=n)
	{
		return 0;
	}

	return sqrt_u32((wt_u32)(wt_u16)n);
}

static wt_u16 sqrt_u16(wt_u16 n)
{
	return sqrt_u32((wt_u32)n);
}

static wt_i64 sqrt_i64(wt_i64 n0)
{
	wt_u64 n = n0, m, y, b;

	if(n0<=0) return 0;

	m = 0x4000000000000000ull; /* TODO: UINT64_C */
	y = 0ul;

	while(m != 0ul)
	{
		b = y | m;
		y = y >> 1;
		if(n >= b)
		{
			n = n - b;
			y = y | m;
		}
		m = m >> 2;
	}
	return y;
}

static wt_u64 sqrt_u64(wt_u64 n)
{
	wt_u64 m, y, b;

	m = 0x4000000000000000ull; /* TODO: UINT64_C */
	y = 0ul;

	while(m != 0ul)
	{
		b = y | m;
		y = y >> 1;
		if(n >= b)
		{
			n = n - b;
			y = y | m;
		}
		m = m >> 2;
	}
	return y;
}

static wt_bool _wtr_strnchr(const char* str, wt_u32 len, const char ch)
{
	while(len--)
	{
		if(ch==*str++)
		{
			return WT_TRUE;
		}
	}

	return WT_FALSE;
}

static wt_u32 _wtr_xorshift(struct wtr_profile* const prof)
{
	wt_u32 t = prof->xs_x ^ (prof->xs_x << 11);
	prof->xs_x = prof->xs_y; prof->xs_y = prof->xs_z; prof->xs_z = prof->xs_w;
	return prof->xs_w = prof->xs_w ^ (prof->xs_w >> 19) ^ t ^ (t >> 8);
}

enum wtr_runresult wtr_mp_undefined(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	((void)prof);
	((void)prog);

	ctx->last_exception = BadProgramAccessException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_conwrite(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_vmptr        vma = *(wt_vmptr*)ctx->dsp;
	struct wtr_str* str;

	CHECK_POP(ctx, sizeof(wt_vmptr));

	WTR_CTX_DS_POP(ctx, wt_vmptr);

	str = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, vma);

	ASSERT_OBJECT_EXISTENCE(str);

	wtr_iocall_writestdout(ctx, str->data, str->length);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_conwritech(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_char ch;

	CHECK_POP(ctx, sizeof(wt_char));
	ch = *(wt_char*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_char);

	*ctx->sbuf = (char)ch;

	wtr_iocall_writestdout(ctx, ctx->sbuf, 1);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_conreadln(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_char ch;

	((void)prog);

	CHECK_POP(ctx, sizeof(wt_char));

	ch = *(wt_char*)ctx->dsp;

	WTR_CTX_DS_POP(ctx, wt_char);

	wtr_iocall_readlnstdin(ctx, (char)ch);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_i8sqrt(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_i32* top = (wt_i32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof((wt_i8)(wt_u8)*top));
	*top = sqrt_i8(*top);
	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_u8sqrt(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32* top = (wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = sqrt_u8((wt_u8)*top);
	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_i16sqrt(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_i32* top = (wt_i32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = sqrt_i16((wt_i16)(wt_u16)*top);
	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_u16sqrt(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32* top = (wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = sqrt_u16((wt_u16)*top);
	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_i32sqrt(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_i32* top = (wt_i32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = sqrt_i32(*top);
	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_u32sqrt(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32* top = (wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = sqrt_u32(*top);
	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_i64sqrt(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_i64* top = (wt_i64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = sqrt_i64(*top);
	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_u64sqrt(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u64* top = (wt_u64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = sqrt_u64(*top);
	return WTRR_RUN;
}

#if WTRUNTIME_ENABLE_FLOAT32

enum wtr_runresult wtr_mp_f32sqrt(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f32* top = (wt_f32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = (wt_f32)sqrt(*top);
	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT32 */

#if WTRUNTIME_ENABLE_FLOAT64

enum wtr_runresult wtr_mp_f64sqrt(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f64* top = (wt_f64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = sqrt(*top);
	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT64 */

#if WTRUNTIME_ENABLE_FLOAT32

enum wtr_runresult wtr_mp_f32log(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f32* top = (wt_f32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = (wt_f32)log(*top);
	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT32 */

#if WTRUNTIME_ENABLE_FLOAT64

enum wtr_runresult wtr_mp_f64log(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f64* top = (wt_f64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = log(*top);
	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT64 */

#if WTRUNTIME_ENABLE_FLOAT32

enum wtr_runresult wtr_mp_f32exp(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f32* top = (wt_f32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = (wt_f32)exp(*top);
	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT32 */

#if WTRUNTIME_ENABLE_FLOAT64

enum wtr_runresult wtr_mp_f64exp(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f64* top = (wt_f64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = exp(*top);
	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT64 */

#if WTRUNTIME_ENABLE_FLOAT32

enum wtr_runresult wtr_mp_f32sin(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f32* top = (wt_f32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = (wt_f32)sin(*top);
	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT32 */

#if WTRUNTIME_ENABLE_FLOAT64

enum wtr_runresult wtr_mp_f64sin(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f64* top = (wt_f64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = sin(*top);
	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT64 */

#if WTRUNTIME_ENABLE_FLOAT32

enum wtr_runresult wtr_mp_f32cos(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f32* top = (wt_f32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = (wt_f32)cos(*top);
	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT32 */

#if WTRUNTIME_ENABLE_FLOAT64

enum wtr_runresult wtr_mp_f64cos(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f64* top = (wt_f64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = cos(*top);
	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT64 */

#if WTRUNTIME_ENABLE_FLOAT32

enum wtr_runresult wtr_mp_f32tan(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f32* top = (wt_f32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = (wt_f32)tan(*top);
	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT32 */

#if WTRUNTIME_ENABLE_FLOAT64

enum wtr_runresult wtr_mp_f64tan(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f64* top = (wt_f64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = tan(*top);
	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT64 */

#if WTRUNTIME_ENABLE_FLOAT32

enum wtr_runresult wtr_mp_f32asin(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f32* top = (wt_f32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = (wt_f32)asin(*top);
	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT32 */

#if WTRUNTIME_ENABLE_FLOAT64

enum wtr_runresult wtr_mp_f64asin(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f64* top = (wt_f64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = asin(*top);
	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT64 */

#if WTRUNTIME_ENABLE_FLOAT32

enum wtr_runresult wtr_mp_f32acos(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f32* top = (wt_f32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = (wt_f32)acos(*top);
	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT32 */

#if WTRUNTIME_ENABLE_FLOAT64

enum wtr_runresult wtr_mp_f64acos(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f64* top = (wt_f64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = acos(*top);
	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT64 */

#if WTRUNTIME_ENABLE_FLOAT32

enum wtr_runresult wtr_mp_f32atan(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f32* top = (wt_f32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = (wt_f32)atan(*top);
	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT32 */

#if WTRUNTIME_ENABLE_FLOAT64

enum wtr_runresult wtr_mp_f64atan(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f64* top = (wt_f64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = atan(*top);
	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT64 */

#if WTRUNTIME_ENABLE_FLOAT32

enum wtr_runresult wtr_mp_f32atan2(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f32  x;
	wt_f32* top = (wt_f32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top)*2);
	x = *top++;
	WTR_CTX_DS_POP(ctx, wt_f32);
	*top = (wt_f32)atan2(*top, x);
	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT32 */

#if WTRUNTIME_ENABLE_FLOAT64

enum wtr_runresult wtr_mp_f64atan2(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f64  x;
	wt_f64* top = (wt_f64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top)*2);
	x = *top++;
	WTR_CTX_DS_POP(ctx, wt_f64);
	*top = atan2(*top, x);
	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT64 */

#if WTRUNTIME_ENABLE_FLOAT32

enum wtr_runresult wtr_mp_f32pow(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f32  exponent;
	wt_f32* top = (wt_f32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top)*2);
	exponent = *top++;
	WTR_CTX_DS_POP(ctx, wt_f32);
	*top = (wt_f32)pow(*top, exponent);
	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT32 */

#if WTRUNTIME_ENABLE_FLOAT64

enum wtr_runresult wtr_mp_f64pow(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f64  exponent;
	wt_f64* top = (wt_f64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top)*2);
	exponent = *top++;
	WTR_CTX_DS_POP(ctx, wt_f64);
	*top = pow(*top, exponent);
	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT64 */

enum wtr_runresult wtr_mp_sarb(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32  amount;
	wt_u32* top = (wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top)*2);
	amount = *top++;
	WTR_CTX_DS_POP(ctx, wt_u32);

	if(amount<=7)
	{
		const wt_u32 shifted_up   = (*top) << 24;
		const wt_i32 shifted_down = ((wt_i32)shifted_up) >> (24 + amount);
		*top = ((wt_u32)shifted_down) & 0xFFu;
		return WTRR_RUN;
	}

	ctx->last_exception = BadProgramAccessException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_sarh(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32  amount;
	wt_u32* top = (wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top)*2);
	amount = *top++;
	WTR_CTX_DS_POP(ctx, wt_u32);

	if(amount<=15)
	{
		const wt_u32 shifted_up   = (*top) << 16;
		const wt_i32 shifted_down = ((wt_i32)shifted_up) >> (16 + amount);
		*top = ((wt_u32)shifted_down) & 0xFFFFu;
		return WTRR_RUN;
	}

	ctx->last_exception = BadProgramAccessException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_negb(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32* top = (wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = (-((wt_i8)*top)) & 0xFF;
	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_negh(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32* top = (wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = (-((wt_i16)*top)) & 0xFFFF;
	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strcmp(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_vmptr        ar;
	wt_vmptr        al;
	struct wtr_str* sr;
	struct wtr_str* sl;
	const char*     psr;
	const char*     psl;

	((void)prof);

	CHECK_POP(ctx, sizeof(wt_vmptr)*2);

	ar = *(wt_vmptr*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_vmptr);
	sr = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, ar);
	al = *(wt_vmptr*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_vmptr);
	sl = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, al);

	psr = sr->data;
	psl = sl->data;

	while(*psl && *psr && *psl==*psr)
	{
		psl++;
		psr++;
	}

	WTR_CTX_DS_PUSH(ctx, wt_i32, *psl - *psr);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_stricmp(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_vmptr        ar;
	wt_vmptr        al;
	struct wtr_str* sr;
	struct wtr_str* sl;
	const char*     psr;
	const char*     psl;

	((void)prof);

	CHECK_POP(ctx, sizeof(wt_vmptr)*2);

	ar = *(wt_vmptr*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_vmptr);
	sr = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, ar);
	al = *(wt_vmptr*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_vmptr);
	sl = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, al);

	psr = sr->data;
	psl = sl->data;

	while(*psl && *psr && tolower(*psl)==tolower(*psr))
	{
		psl++;
		psr++;
	}

	WTR_CTX_DS_PUSH(ctx, wt_i32, *psl - *psr);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strdup(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	struct wtr_str* sr;
	struct wtr_str* res;

	CHECK_POP(ctx, sizeof(wt_vmptr));

	sr = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);

	if(NULL!=(res=wtr_str_create(prog, sr->data, sr->length, WT_FALSE)))
	{
		*(wt_vmptr*)ctx->dsp = WTR_OBJ_TO_VMADDR(prog, res);
		STR_REGISTER(prof, prog, ctx, res, *(wt_vmptr*)ctx->dsp);
	}
	else
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strcat(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_vmptr        ar;
	wt_vmptr        al;
	struct wtr_str* sr;
	struct wtr_str* sl;
	struct wtr_str* res;

	CHECK_POP(ctx, sizeof(wt_vmptr)*2);

	ar = *(wt_vmptr*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_vmptr);
	sr = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, ar);
	al = *(wt_vmptr*)ctx->dsp;
	sl = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, al);

	if(NULL==(res=wtr_str_create(prog, NULL, sl->length + sr->length, WT_FALSE)))
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	memcpy(res->data, sl->data, sl->length);
	memcpy(res->data+sl->length, sr->data, sr->length);

	*(wt_vmptr*)ctx->dsp = WTR_OBJ_TO_VMADDR(prog, res);
	STR_REGISTER(prof, prog, ctx, res, *(wt_vmptr*)ctx->dsp);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strrev(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	struct wtr_str* sr;
	struct wtr_str* res;

	CHECK_POP(ctx, sizeof(wt_vmptr));

	sr = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);

	if(NULL!=(res=wtr_str_create(prog, NULL, sr->length, WT_FALSE)))
	{
		const char* psr  = sr->data + sr->length - 1;
		char*       pres = res->data;
		char*       end  = res->data + res->length;

		while(pres<end)
		{
			*pres++ = *psr--;
		}

		*(wt_vmptr*)ctx->dsp = WTR_OBJ_TO_VMADDR(prog, res);
		STR_REGISTER(prof, prog, ctx, res, *(wt_vmptr*)ctx->dsp);
	}
	else
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strsub(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32          offset;
	wt_u32          count;
	struct wtr_str* sr;
	struct wtr_str* res;

	CHECK_POP(ctx, sizeof(wt_vmptr)+sizeof(wt_u32)+sizeof(wt_u32));

	count = *(wt_u32*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_u32);
	offset = *(wt_u32*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_u32);
	sr = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);

	if(offset>=sr->length)
	{
		offset = sr->length;
		count  = 0;
	}
	else
	if(offset+count>sr->length)
	{
		count  = sr->length - offset;
	}

	if((res=wtr_str_create(prog, sr->data+offset, count, WT_FALSE))!=NULL)
	{
		*(wt_vmptr*)ctx->dsp = WTR_OBJ_TO_VMADDR(prog, res);
		STR_REGISTER(prof, prog, ctx, res, *(wt_vmptr*)ctx->dsp);
	}
	else
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strsub2(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32          offset;
	wt_u32          count;
	struct wtr_str* sr;
	struct wtr_str* res;

	CHECK_POP(ctx, sizeof(wt_vmptr)+sizeof(wt_u32));

	offset = *(wt_u32*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_u32);
	sr = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	count = sr->length - offset;

	if(offset>=sr->length)
	{
		offset = sr->length;
		count  = 0;
	}

	if((res=wtr_str_create(prog, sr->data+offset, count, WT_FALSE))!=NULL)
	{
		*(wt_vmptr*)ctx->dsp = WTR_OBJ_TO_VMADDR(prog, res);
		STR_REGISTER(prof, prog, ctx, res, *(wt_vmptr*)ctx->dsp);
	}
	else
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strlen(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_vmptr        ar;
	struct wtr_str* sr;

	((void)prof);

	CHECK_POP(ctx, sizeof(wt_vmptr));

	ar = *(wt_vmptr*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_vmptr);
	sr = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, ar);

	WTR_CTX_DS_PUSH(ctx, wt_u32, sr->length);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strupper(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	struct wtr_str* sr;
	struct wtr_str* res;

	CHECK_POP(ctx, sizeof(wt_vmptr));

	sr = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);

	if((res=wtr_str_create(prog, NULL, sr->length, WT_FALSE))!=NULL)
	{
		const char* psr  = sr->data;
		char*       pres = res->data;
		char*       end  = res->data + res->length;

		while(pres<end)
		{
			*pres++ = toupper(*psr++);
		}

		*(wt_vmptr*)ctx->dsp = WTR_OBJ_TO_VMADDR(prog, res);
		STR_REGISTER(prof, prog, ctx, res, *(wt_vmptr*)ctx->dsp);
	}
	else
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strlower(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	struct wtr_str* sr;
	struct wtr_str* res;

	CHECK_POP(ctx, sizeof(wt_vmptr));

	sr = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);

	if((res=wtr_str_create(prog, NULL, sr->length, WT_FALSE))!=NULL)
	{
		const char* psr  = sr->data;
		char*       pres = res->data;
		char*       end  = res->data + res->length;

		while(pres<end)
		{
			*pres++ = tolower(*psr++);
		}

		*(wt_vmptr*)ctx->dsp = WTR_OBJ_TO_VMADDR(prog, res);
		STR_REGISTER(prof, prog, ctx, res, *(wt_vmptr*)ctx->dsp);
	}
	else
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strtrim(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	struct wtr_str* sr;
	struct wtr_str* res;
	char*           s;
	char*           s2;
	wt_u32          len;

	CHECK_POP(ctx, sizeof(wt_vmptr));

	sr  = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	s   = sr->data;
	len = sr->length;

	while(*s && '\x20'>=*s)
	{
		--len;
		++s;
	}

	s2 = s + len - 1;

	while(len && *s2 && '\x20'>=*s2)
	{
		--len;
		--s2;
	}

	if((res=wtr_str_create(prog, s, len, WT_FALSE))!=NULL)
	{
		*(wt_vmptr*)ctx->dsp = WTR_OBJ_TO_VMADDR(prog, res);
		STR_REGISTER(prof, prog, ctx, res, *(wt_vmptr*)ctx->dsp);
	}
	else
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strtrim2(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_vmptr        aset;
	struct wtr_str* sset;
	struct wtr_str* sr;
	struct wtr_str* res;
	char*           s;
	char*           s2;
	wt_u32          len;

	CHECK_POP(ctx, sizeof(wt_vmptr)*2);

	aset = *(wt_vmptr*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_vmptr);
	sset = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, aset);

	sr  = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	s   = sr->data;
	len = sr->length;

	while(_wtr_strnchr(sset->data, sset->length, *s))
	{
		--len;
		++s;
	}

	s2 = s + len - 1;

	while(len && _wtr_strnchr(sset->data, sset->length, *s2))
	{
		--len;
		--s2;
	}

	if((res=wtr_str_create(prog, s, len, WT_FALSE))!=NULL)
	{
		*(wt_vmptr*)ctx->dsp = WTR_OBJ_TO_VMADDR(prog, res);
		STR_REGISTER(prof, prog, ctx, res, *(wt_vmptr*)ctx->dsp);
	}
	else
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strempty(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	struct wtr_str* sr;

	CHECK_POP(ctx, sizeof(wt_vmptr));

	sr = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);

	ctx->dsp += sizeof(wt_vmptr) - sizeof(wt_bool);
	*(wt_bool*)ctx->dsp = (0==sr->length);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strstartswith(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	struct wtr_str* sl;
	struct wtr_str* sr;
	wt_bool         result;

	CHECK_POP(ctx, sizeof(wt_vmptr)*2);

	sr = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);
	sl = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);

	if(sl->length<sr->length)
	{
		result = WT_FALSE;
	}
	else
	{
		result = (0==memcmp(sl->data, sr->data, sr->length));
	}

	WTR_CTX_DS_PUSH(ctx, wt_bool, result);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strendswith(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	struct wtr_str* sl;
	struct wtr_str* sr;
	wt_bool         result;

	CHECK_POP(ctx, sizeof(wt_vmptr)*2);

	sr = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);
	sl = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);

	if(sl->length<sr->length)
	{
		result = WT_FALSE;
	}
	else
	{
		result = (0==memcmp(sl->data+sl->length-sr->length, sr->data, sr->length));
	}

	WTR_CTX_DS_PUSH(ctx, wt_bool, result);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strchars(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	char            ch;
	wt_u32          n;
	struct wtr_str* res;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_char) + sizeof(wt_u32));

	n = (*(wt_u32*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_u32);
	ch = (char)(*(wt_char*)ctx->dsp);
	ctx->dsp -= (sizeof(wt_vmptr) - sizeof(wt_char));

	if((res=wtr_str_create(prog, NULL, n, WT_FALSE))!=NULL)
	{
		char* p = res->data;

		while(n--)
		{
			*p++ = ch;
		}

		*(wt_vmptr*)ctx->dsp = WTR_OBJ_TO_VMADDR(prog, res);
		STR_REGISTER(prof, prog, ctx, res, *(wt_vmptr*)ctx->dsp);
	}
	else
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strchars2(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	char            ch;
	wt_u32          n;
	struct wtr_str* s;
	struct wtr_str* res;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_char) + sizeof(wt_u32) + sizeof(wt_vmptr));

	s = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);
	n = (*(wt_u32*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_u32);
	ch = (char)(*(wt_char*)ctx->dsp);
	ctx->dsp -= (sizeof(wt_vmptr) - sizeof(wt_char));

	if((res=wtr_str_create(prog, NULL, n + s->length, WT_FALSE))!=NULL)
	{
		char* p = res->data;

		while(n--)
		{
			*p++ = ch;
		}
		memcpy(p, s->data, s->length);

		*(wt_vmptr*)ctx->dsp = WTR_OBJ_TO_VMADDR(prog, res);
		STR_REGISTER(prof, prog, ctx, res, *(wt_vmptr*)ctx->dsp);
	}
	else
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strcopies(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32          n;
	struct wtr_str* s;
	struct wtr_str* res;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_vmptr) + sizeof(wt_u32));

	n = (*(wt_u32*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_u32);
	s = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);

	if((res=wtr_str_create(prog, NULL, n * s->length, WT_FALSE))!=NULL)
	{
		char* p = res->data;

		while(n--)
		{
			memcpy(p, s->data, s->length);
			p += s->length;
		}

		*(wt_vmptr*)ctx->dsp = WTR_OBJ_TO_VMADDR(prog, res);
		STR_REGISTER(prof, prog, ctx, res, *(wt_vmptr*)ctx->dsp);
	}
	else
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strcopies2(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32          n;
	struct wtr_str* s;
	struct wtr_str* suf;
	struct wtr_str* res;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_vmptr)*2 + sizeof(wt_u32));

	suf = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);
	n = (*(wt_u32*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_u32);
	s = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);

	if((res=wtr_str_create(prog, NULL, n * s->length + suf->length, WT_FALSE))!=NULL)
	{
		char* p = res->data;

		while(n--)
		{
			memcpy(p, s->data, s->length);
			p += s->length;
		}

		memcpy(p, suf->data, suf->length);

		*(wt_vmptr*)ctx->dsp = WTR_OBJ_TO_VMADDR(prog, res);
		STR_REGISTER(prof, prog, ctx, res, *(wt_vmptr*)ctx->dsp);
	}
	else
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strcpy(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	struct wtr_str* sdst;
	wt_u32          odst;
	struct wtr_str* ssrc;
	wt_u32          osrc;
	wt_u32          len;

	CHECK_POP(ctx, sizeof(wt_vmptr)*2 + sizeof(wt_u32)*3);

	len = (*(wt_u32*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_u32);
	osrc = (*(wt_u32*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_u32);
	ssrc = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);
	odst = (*(wt_u32*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_u32);
	sdst = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);

	if(odst+len<=sdst->length && osrc+len<=ssrc->length)
	{
		memmove(sdst->data+odst, ssrc->data+osrc, len);
	}
	else
	{
		ctx->last_exception = IndexOutOfRangeException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strstr(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	struct wtr_str* sl;
	struct wtr_str* sr;
	char*           p;

	CHECK_POP(ctx, sizeof(wt_vmptr)*2);

	sr = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);
	sl = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);

	if(NULL!=(p=strstr(sl->data, sr->data)))
	{
		WTR_CTX_DS_PUSH(ctx, wt_u32, p - sl->data);
	}
	else
	{
		WTR_CTX_DS_PUSH(ctx, wt_u32, sl->length);
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strstr2(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	struct wtr_str* sl;
	struct wtr_str* sr;
	char*           p;
	wt_u32          offs;

	CHECK_POP(ctx, sizeof(wt_vmptr)*2);

	sr = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);
	offs = (*(wt_u32*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_u32);
	sl = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);

	if(offs<=sl->length)
	{
		if(NULL!=(p=strstr(sl->data+offs, sr->data)))
		{
			WTR_CTX_DS_PUSH(ctx, wt_u32, p - sl->data);
		}
		else
		{
			WTR_CTX_DS_PUSH(ctx, wt_u32, sl->length);
		}
	}
	else
	{
		ctx->last_exception = IndexOutOfRangeException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strrstr(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	struct wtr_str* sl;
	struct wtr_str* sr;
	wt_u32          result;
	char*           p;

	CHECK_POP(ctx, sizeof(wt_vmptr)*2);

	sr = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);
	sl = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);

	if(NULL!=(p=strstr(sl->data, sr->data)))
	{
		do
		{
			result = p - sl->data;
			p = strstr(p+1, sr->data);
		}
		while(p);

		WTR_CTX_DS_PUSH(ctx, wt_u32, result);
	}
	else
	{
		WTR_CTX_DS_PUSH(ctx, wt_u32, sl->length);
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strrstr2(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	struct wtr_str* sl;
	struct wtr_str* sr;
	wt_u32          offs;
	wt_u32          result;
	char*           p;

	CHECK_POP(ctx, sizeof(wt_vmptr)*2);

	sr = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);
	offs = (*(wt_u32*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_u32);
	sl = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);

	if(offs<=sl->length)
	{
		if(NULL!=(p=strstr(sl->data+offs, sr->data)))
		{
			do
			{
				result = p - sl->data;
				p = strstr(p+1, sr->data);
			}
			while(p);

			WTR_CTX_DS_PUSH(ctx, wt_u32, result);
		}
		else
		{
			WTR_CTX_DS_PUSH(ctx, wt_u32, sl->length);
		}
	}
	else
	{
		ctx->last_exception = IndexOutOfRangeException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strchr(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	char            ch;
	struct wtr_str* s;
	char*           p;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_char) + sizeof(wt_vmptr));

	ch = (char)(*(wt_char*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_char);
	s = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);

	if(NULL!=(p=strchr(s->data, ch)))
	{
		WTR_CTX_DS_PUSH(ctx, wt_u32, p - s->data);
	}
	else
	{
		WTR_CTX_DS_PUSH(ctx, wt_u32, s->length);
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strchr2(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	char            ch;
	wt_u32          offs;
	struct wtr_str* s;
	char*           p;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_char) + sizeof(wt_u32) + sizeof(wt_vmptr));

	ch = (char)(*(wt_char*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_char);
	offs = (*(wt_u32*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_u32);
	s = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);

	if(offs<=s->length)
	{
		if(NULL!=(p=strchr(s->data+offs, ch)))
		{
			WTR_CTX_DS_PUSH(ctx, wt_u32, p - s->data);
		}
		else
		{
			WTR_CTX_DS_PUSH(ctx, wt_u32, s->length);
		}
	}
	else
	{
		ctx->last_exception = IndexOutOfRangeException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strrchr(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	char            ch;
	struct wtr_str* s;
	char*           p;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_char) + sizeof(wt_vmptr));

	ch = (char)(*(wt_char*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_char);
	s = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);

	if(NULL!=(p=strrchr(s->data, ch)))
	{
		WTR_CTX_DS_PUSH(ctx, wt_u32, p - s->data);
	}
	else
	{
		WTR_CTX_DS_PUSH(ctx, wt_u32, s->length);
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strrchr2(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	char            ch;
	wt_u32          offs;
	struct wtr_str* s;
	char*           p;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_char) + sizeof(wt_u32) + sizeof(wt_vmptr));

	ch = (char)(*(wt_char*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_char);
	offs = (*(wt_u32*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_u32);
	s = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);

	if(offs<=s->length)
	{
		if(NULL!=(p=strrchr(s->data+offs, ch)))
		{
			WTR_CTX_DS_PUSH(ctx, wt_u32, p - s->data);
		}
		else
		{
			WTR_CTX_DS_PUSH(ctx, wt_u32, s->length);
		}
	}
	else
	{
		ctx->last_exception = IndexOutOfRangeException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strspn(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	struct wtr_str* sl;
	struct wtr_str* sr;
	wt_u32          result;

	CHECK_POP(ctx, sizeof(wt_vmptr)*2);

	sr = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);
	sl = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);

	result = (wt_u32)strspn(sl->data, sr->data);
	WTR_CTX_DS_PUSH(ctx, wt_u32, result);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strspn2(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	struct wtr_str* sl;
	struct wtr_str* sr;
	wt_u32          offs;
	wt_u32          result;

	CHECK_POP(ctx, sizeof(wt_vmptr)*2 + sizeof(wt_u32));

	sr = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);
	offs = (*(wt_u32*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_u32);
	sl = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);

	if(offs<=sl->length)
	{
		result = (wt_u32)strspn(sl->data+offs, sr->data);
		WTR_CTX_DS_PUSH(ctx, wt_u32, offs+result);
	}
	else
	{
		ctx->last_exception = IndexOutOfRangeException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strcspn(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	struct wtr_str* sl;
	struct wtr_str* sr;
	wt_u32          result;

	CHECK_POP(ctx, sizeof(wt_vmptr)*2);

	sr = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);
	sl = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);

	result = (wt_u32)strcspn(sl->data, sr->data);
	WTR_CTX_DS_PUSH(ctx, wt_u32, result);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strcspn2(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	struct wtr_str* sl;
	struct wtr_str* sr;
	wt_u32          offs;
	wt_u32          result;

	CHECK_POP(ctx, sizeof(wt_vmptr)*2 + sizeof(wt_u32));

	sr = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);
	offs = (*(wt_u32*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_u32);
	sl = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);

	if(offs<=sl->length)
	{
		result = (wt_u32)strcspn(sl->data+offs, sr->data);
		WTR_CTX_DS_PUSH(ctx, wt_u32, offs+result);
	}
	else
	{
		ctx->last_exception = IndexOutOfRangeException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strfromutf8(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	struct wtr_str* sr;
	struct wtr_str* res;
	char*           buf;
	wt_u8*          ptr;
	wt_u8*          end;
	wt_u32          len;

	CHECK_POP(ctx, sizeof(wt_vmptr));

	sr  = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	ptr = (wt_u8*)sr->data;
	end = ptr + sr->length;
	buf = (char*)wtr_program_malloc(prog, sr->length + 1);
	len = 0;

	if(!buf)
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	while(ptr<end)
	{
		const wt_u8 ub   = *ptr++;
		wt_u32      code = 0u;
		wt_u32      more = 0u;

		if(ub < UTF8_LAST_1)
		{
			buf[len++] = (char)ub;
			continue;
		}
		else
		if((ub & ~UTF8_SEQ2_DATA_MASK) == UTF8_SEQ2)
		{
			code = ub & UTF8_SEQ2_DATA_MASK;
			more = 1;
		}
		else
		if((ub & ~UTF8_SEQ3_DATA_MASK) == UTF8_SEQ3)
		{
			code = ub & UTF8_SEQ3_DATA_MASK;
			more = 2;
		}
		else
		if((ub & ~UTF8_SEQ4_DATA_MASK) == UTF8_SEQ4)
		{
			code = ub & UTF8_SEQ4_DATA_MASK;
			more = 3;
		}
		else
		{
free_and_raise:
			wtr_program_free(prog, buf);
			ctx->last_exception = UnicodeErrorException;
			return WTRR_RAISE;
		}

		if(ptr+more>end)
		{
			goto free_and_raise;
		}

		while(more--)
		{
			code <<= UTF8_CSEQ_BITS;
			code |= (*ptr++ & UTF8_CSEQ_DATA_MASK); 
		}

		if(code>255)
		{
			goto free_and_raise;
		}

		buf[len++] = (char)(wt_u8)code;
	}

	buf[len] = '\0';

	res = wtr_str_create(prog, buf, len, WT_FALSE);
	wtr_program_free(prog, buf);

	if(res)
	{
		*(wt_vmptr*)ctx->dsp = WTR_OBJ_TO_VMADDR(prog, res);
		STR_REGISTER(prof, prog, ctx, res, *(wt_vmptr*)ctx->dsp);
	}
	else
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strtoutf8(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	struct wtr_str* sr;
	struct wtr_str* res;
	wt_u8*          buf;
	wt_u8*          ptr;
	wt_u8*          end;
	wt_u32          len;

	CHECK_POP(ctx, sizeof(wt_vmptr));

	sr  = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	ptr = (wt_u8*)sr->data;
	end = ptr + sr->length;
	buf = (wt_u8*)wtr_program_malloc(prog, sr->length*2 + 1);
	len = 0;

	if(NULL==buf)
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	while(ptr<end)
	{
		wt_u8 ub = *ptr++;

		if(ub < UTF8_LAST_1)
		{
			buf[len++] = ub;
		}
		else
		{
			buf[len+1] = UTF8_CSEQ | (ub & UTF8_CSEQ_DATA_MASK);
			ub >>= UTF8_CSEQ_BITS;
			buf[len+0] = UTF8_SEQ2 | ub;

			len += 2;
		}
	}

	buf[len] = '\0';

	res = wtr_str_create(prog, (char*)buf, len, WT_FALSE);
	wtr_program_free(prog, buf);

	if(res)
	{
		*(wt_vmptr*)ctx->dsp = WTR_OBJ_TO_VMADDR(prog, res);
		STR_REGISTER(prof, prog, ctx, res, *(wt_vmptr*)ctx->dsp);
	}
	else
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_chartostring(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	char            ch;
	struct wtr_str* res;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_char));
	CHECK_PUSH(ctx, sizeof(wt_vmptr));

	ch = (char)(*(wt_char*)ctx->dsp);
	ctx->dsp -= (sizeof(wt_vmptr) - sizeof(wt_char));

	if((res=wtr_str_create(prog, &ch, 1, WT_FALSE))!=NULL)
	{
		*(wt_vmptr*)ctx->dsp = WTR_OBJ_TO_VMADDR(prog, res);
		STR_REGISTER(prof, prog, ctx, res, *(wt_vmptr*)ctx->dsp);
	}
	else
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_charcmp(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_char lhs;
	wt_char rhs;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_char)*2);

	rhs = *(wt_char*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_char);
	lhs = *(wt_char*)ctx->dsp;

	ctx->dsp -= (sizeof(wt_i32) - sizeof(wt_char));

	*(wt_i32*)ctx->dsp = lhs - rhs;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_charadd(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_char rhs;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_char)*2);

	rhs = *(wt_char*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_char);

	*(wt_char*)ctx->dsp = (*(wt_char*)ctx->dsp + rhs) & 0xff;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_charsub(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_char rhs;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_char)*2);

	rhs = *(wt_char*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_char);

	*(wt_char*)ctx->dsp = (*(wt_char*)ctx->dsp - rhs) & 0xff;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_charinc(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_i32 rhs;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_char)*2);

	rhs = *(wt_i32*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_i32);

	*(wt_char*)ctx->dsp = (*(wt_char*)ctx->dsp + rhs) & 0xff;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_chardiff(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_char lhs;
	wt_char rhs;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_char)*2);

	rhs = *(wt_char*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_char);
	lhs = *(wt_char*)ctx->dsp;

	ctx->dsp -= (sizeof(wt_i32) - sizeof(wt_char));

	*(wt_i32*)ctx->dsp = lhs - rhs;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_chartolower(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_char ch;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_char));

	ch = *(wt_char*)ctx->dsp;

	if('A'<=((char)ch) && ((char)ch)<='Z')
	{
		*(wt_char*)ctx->dsp = ch - 'A' + 'a';
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_chartoupper(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_char ch;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_char));

	ch = *(wt_char*)ctx->dsp;

	if('a'<=((char)ch) && ((char)ch)<='z')
	{
		*(wt_char*)ctx->dsp = ch - 'a' + 'A';
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_seqempty(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	struct wtr_seq* sr;

	CHECK_POP(ctx, sizeof(wt_vmptr));

	sr = (struct wtr_seq*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);

	ctx->dsp += sizeof(wt_vmptr) - sizeof(wt_bool);
	*(wt_bool*)ctx->dsp = (0==sr->length);

	return WTRR_RUN;
}

#define _CONVERT_X_TO_STRING(_type_) \
enum wtr_runresult wtr_mp_c##_type_##string(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx) \
{ \
	wt_##_type_     rhs; \
	struct wtr_str* res; \
	char            buf[64]; \
\
	CHECK_POP(ctx, sizeof(wt_##_type_)); \
\
	rhs = *(wt_##_type_*)ctx->dsp; \
	ctx->dsp -= (sizeof(wt_vmptr) - sizeof(wt_##_type_)); \
\
	if((res=wtr_str_create(prog, wtr_convert_##_type_##_to_dec(rhs, buf, buf+sizeof(buf)), 0, WT_TRUE))!=NULL) \
	{ \
		*(wt_vmptr*)ctx->dsp = WTR_OBJ_TO_VMADDR(prog, res); \
		STR_REGISTER(prof, prog, ctx, res, *(wt_vmptr*)ctx->dsp); \
	} \
	else \
	{ \
		ctx->last_exception = OutOfMemoryException; \
		return WTRR_RAISE; \
	} \
\
	return WTRR_RUN; \
} \

_CONVERT_X_TO_STRING(i32)
_CONVERT_X_TO_STRING(u32)
_CONVERT_X_TO_STRING(i64)
_CONVERT_X_TO_STRING(u64)
#if WTRUNTIME_ENABLE_FLOAT32
_CONVERT_X_TO_STRING(f32)
#endif /* WTRUNTIME_ENABLE_FLOAT32 */
#if WTRUNTIME_ENABLE_FLOAT64
_CONVERT_X_TO_STRING(f64)
#endif /* WTRUNTIME_ENABLE_FLOAT64 */
_CONVERT_X_TO_STRING(bool)

#define _CONVERT_STRING_TO_X(_type_) \
enum wtr_runresult wtr_mp_cstring##_type_(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx) \
{ \
	struct wtr_str* sr; \
\
	((void)prof); \
	((void)prog); \
\
	CHECK_POP(ctx, sizeof(wt_vmptr)); \
	CHECK_PUSH(ctx, sizeof(wt_##_type_)); \
\
	sr = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp); \
	ctx->dsp -= (sizeof(wt_##_type_) - sizeof(wt_vmptr)); \
\
	*(wt_##_type_*)ctx->dsp = wtr_convert_dec_to_##_type_(sr->data); \
\
	return WTRR_RUN; \
} \

_CONVERT_STRING_TO_X(i32)
_CONVERT_STRING_TO_X(u32)
_CONVERT_STRING_TO_X(i64)
_CONVERT_STRING_TO_X(u64)
#if WTRUNTIME_ENABLE_FLOAT32
_CONVERT_STRING_TO_X(f32)
#endif /* WTRUNTIME_ENABLE_FLOAT32 */
#if WTRUNTIME_ENABLE_FLOAT64
_CONVERT_STRING_TO_X(f64)
#endif /* WTRUNTIME_ENABLE_FLOAT64 */
_CONVERT_STRING_TO_X(bool)

enum wtr_runresult wtr_mp_ci32u32(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32 u = *(wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));

	if(u>0x7fffffffU)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cu32i32(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32 u = *(wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));

	if(u>0x7fffffffU)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cu32u64(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32 u = *(wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));
	CHECK_PUSH(ctx, sizeof(wt_u64));

	ctx->dsp -= sizeof(wt_u64) - sizeof(wt_u32);
	*(wt_u64*)ctx->dsp = (wt_u64)u;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cu32i64(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32 u = *(wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));
	CHECK_PUSH(ctx, sizeof(wt_u64));

	ctx->dsp -= sizeof(wt_u64) - sizeof(wt_u32);
	*(wt_u64*)ctx->dsp = (wt_u64)u;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_ci32u64(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32 u = *(wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));
	CHECK_PUSH(ctx, sizeof(wt_u64));

	if(u>0x7fffffffU)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	ctx->dsp -= sizeof(wt_u64) - sizeof(wt_u32);
	*(wt_u64*)ctx->dsp = (wt_u64)u;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_ci32i64(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_i32 i = *(wt_i32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));
	CHECK_PUSH(ctx, sizeof(wt_u32));

	ctx->dsp -= sizeof(wt_i64) - sizeof(wt_i32);
	*(wt_i64*)ctx->dsp = (wt_i64)i;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cu64u32(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u64 u = *(wt_u64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));

	if(u>0xffffffffULL)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	ctx->dsp += sizeof(wt_u64) - sizeof(wt_u32);
	*(wt_u32*)ctx->dsp = (wt_u32)u;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cu64i32(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u64 u = *(wt_u64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));

	if(u>0x7fffffffULL)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	ctx->dsp += sizeof(wt_u64) - sizeof(wt_u32);
	*(wt_u32*)ctx->dsp = (wt_u32)u;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_ci64u32(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u64 u = *(wt_u64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));

	if(u>0xffffffffULL)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	ctx->dsp += sizeof(wt_u64) - sizeof(wt_u32);
	*(wt_u32*)ctx->dsp = (wt_u32)u;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_ci64i32(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_i64 u = *(wt_i64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));

	if(u>0x7fffffffLL || u<-0x80000000LL)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	ctx->dsp += sizeof(wt_u64) - sizeof(wt_u32);
	*(wt_i32*)ctx->dsp = (wt_i32)u;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_ci64u64(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u64 u = *(wt_u64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));

	if(u>0x7fffffffffffffffULL)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cu64i64(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u64 u = *(wt_u64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));

	if(u>0x7fffffffffffffffULL)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

#if WTRUNTIME_ENABLE_FLOAT32

enum wtr_runresult wtr_mp_cf32i32(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f32 f = *(wt_f32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));

	if(f<-2147483648.f || f>2147483647.f)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	*(wt_i32*)ctx->dsp = (wt_i32)f;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cf32u32(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f32 f = *(wt_f32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));

	if(f<0 || f>0xffffffff)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	*(wt_u32*)ctx->dsp = (wt_u32)f;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_ci32f32(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_i32 i = *(wt_i32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));

	*(wt_f32*)ctx->dsp = (wt_f32)i;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cu32f32(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32 u = *(wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));

	*(wt_f32*)ctx->dsp = (wt_f32)u;

	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT32 */

#if WTRUNTIME_ENABLE_FLOAT64

enum wtr_runresult wtr_mp_ci32f64(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_i32 i = *(wt_i32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));
	CHECK_PUSH(ctx, sizeof(wt_u32));

	ctx->dsp -= sizeof(wt_f64) - sizeof(wt_i32);
	*(wt_f64*)ctx->dsp = (wt_f64)i;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cu32f64(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32 u = *(wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));
	CHECK_PUSH(ctx, sizeof(wt_u32));

	ctx->dsp -= sizeof(wt_f64) - sizeof(wt_u32);
	*(wt_f64*)ctx->dsp = (wt_f64)u;

	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT64 */

#if WTRUNTIME_ENABLE_FLOAT32 && WTRUNTIME_ENABLE_FLOAT64

enum wtr_runresult wtr_mp_cf32f64(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f32 f = *(wt_f32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));
	CHECK_PUSH(ctx, sizeof(wt_u32));

	ctx->dsp -= sizeof(wt_f64) - sizeof(wt_f32);
	*(wt_f64*)ctx->dsp = (wt_f64)f;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cf64f32(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f64 f = *(wt_f64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));

	if(f<FLT_MIN || f>FLT_MAX)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	ctx->dsp += sizeof(wt_f64) - sizeof(wt_f32);
	*(wt_f32*)ctx->dsp = (wt_f32)f;

	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT32 && WTRUNTIME_ENABLE_FLOAT64 */

#if WTRUNTIME_ENABLE_FLOAT32

enum wtr_runresult wtr_mp_ci64f32(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_i64 i = *(wt_i64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));

	ctx->dsp += sizeof(wt_i64) - sizeof(wt_f32);
	*(wt_f32*)ctx->dsp = (wt_f32)i;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cu64f32(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u64 u = *(wt_u64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));

	ctx->dsp += sizeof(wt_u64) - sizeof(wt_f32);
	*(wt_f32*)ctx->dsp = (wt_f32)u;

	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT32 */

#if WTRUNTIME_ENABLE_FLOAT64

enum wtr_runresult wtr_mp_ci64f64(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_i64 i = *(wt_i64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));

	*(wt_f64*)ctx->dsp = (wt_f64)i;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cu64f64(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u64 u = *(wt_u64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));

	*(wt_f64*)ctx->dsp = (wt_f64)u;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cf64i32(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f64 f = *(wt_f64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));

	if(f<-2147483648.0 || f>2147483647.0)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	ctx->dsp += sizeof(wt_f64) - sizeof(wt_i32);
	*(wt_i32*)ctx->dsp = (wt_i32)f;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cf64u32(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f64 f = *(wt_f64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));

	if(f<0 || f>0xffffffff)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	ctx->dsp += sizeof(wt_f64) - sizeof(wt_u32);
	*(wt_u32*)ctx->dsp = (wt_u32)f;

	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT64 */

#if WTRUNTIME_ENABLE_FLOAT32

enum wtr_runresult wtr_mp_cf32i64(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f32 f = *(wt_f32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));
	CHECK_PUSH(ctx, sizeof(wt_u32));

	if(f<(wt_f32)(-0x7fffffffffffffff-1) || f>(wt_f32)0x7fffffffffffffff)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	ctx->dsp -= sizeof(wt_i64) - sizeof(wt_i32);
	*(wt_i64*)ctx->dsp = (wt_i64)f;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cf32u64(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f32 f = *(wt_f32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_f32));
	CHECK_PUSH(ctx, sizeof(wt_u32));

	if(f<0.f || f>(wt_f32)0xffffffffffffffff)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	ctx->dsp -= sizeof(wt_u64) - sizeof(wt_f32);
	*(wt_u64*)ctx->dsp = (wt_u64)f;

	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT32 */

#if WTRUNTIME_ENABLE_FLOAT64

enum wtr_runresult wtr_mp_cf64i64(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f64 f = *(wt_f64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));

	if(f<(wt_f64)(-0x7fffffffffffffff-1) || f>(wt_f64)0x7fffffffffffffff)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	*(wt_i64*)ctx->dsp = (wt_i64)f;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cf64u64(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_f64 f = *(wt_f64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));

	if(f<0.0 || f>(wt_f64)0xffffffffffffffff)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	*(wt_u64*)ctx->dsp = (wt_u64)f;

	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT64 */

enum wtr_runresult wtr_mp_cchari32(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_char value = *(wt_char*)ctx->dsp;

	CHECK_POP(ctx, sizeof(value));
	CHECK_PUSH(ctx, sizeof(wt_i32)-sizeof(value));

	ctx->dsp -= sizeof(wt_i32) - sizeof(value);
	*(wt_i32*)ctx->dsp = (wt_i32)value;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_ccharu32(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_char value = *(wt_char*)ctx->dsp;

	CHECK_POP(ctx, sizeof(value));
	CHECK_PUSH(ctx, sizeof(wt_u32)-sizeof(value));

	ctx->dsp -= sizeof(wt_u32) - sizeof(value);
	*(wt_u32*)ctx->dsp = (wt_u32)value;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cchari64(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_char value = *(wt_char*)ctx->dsp;

	CHECK_POP(ctx, sizeof(value));
	CHECK_PUSH(ctx, sizeof(wt_i64)-sizeof(value));

	ctx->dsp -= sizeof(wt_i64) - sizeof(value);
	*(wt_i64*)ctx->dsp = (wt_i64)value;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_ccharu64(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_char value = *(wt_char*)ctx->dsp;

	CHECK_POP(ctx, sizeof(value));
	CHECK_PUSH(ctx, sizeof(wt_u64)-sizeof(value));

	ctx->dsp -= sizeof(wt_u64) - sizeof(value);
	*(wt_u64*)ctx->dsp = (wt_u64)value;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_ci32char(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_i32 value = *(wt_i32*)ctx->dsp;

	CHECK_POP(ctx, sizeof(value));

	if(value<0 || value>255)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	ctx->dsp -= sizeof(value) - sizeof(wt_char);
	*(wt_char*)ctx->dsp = (wt_char)value;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cu32char(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32 value = *(wt_u32*)ctx->dsp;

	CHECK_POP(ctx, sizeof(value));

	if(value>255)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	ctx->dsp -= sizeof(value) - sizeof(wt_char);
	*(wt_char*)ctx->dsp = (wt_char)value;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_ci64char(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_i64 value = *(wt_i64*)ctx->dsp;

	CHECK_POP(ctx, sizeof(value));

	if(value<0 || value>255)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	ctx->dsp -= sizeof(value) - sizeof(wt_char);
	*(wt_char*)ctx->dsp = (wt_char)value;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cu64char(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u64 value = *(wt_u64*)ctx->dsp;

	CHECK_POP(ctx, sizeof(value));

	if(value>255)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	ctx->dsp -= sizeof(value) - sizeof(wt_char);
	*(wt_char*)ctx->dsp = (wt_char)value;

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_ci32toi8(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_i32 u = *(wt_i32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));

	if((-128)<=u && u<=127)
	{
		*(wt_u32*)ctx->dsp &= 0xFFu;
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_ci32tou8(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_i32 u = *(wt_i32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));

	if(0<=u && u<=255)
	{
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_ci32toi16(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_i32 u = *(wt_i32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));

	if((-32768)<=u && u<=32767)
	{
		*(wt_u32*)ctx->dsp &= 0xFFFFu;
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_ci32tou16(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_i32 u = *(wt_i32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));

	if(0<=u && u<=65535)
	{
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_cu32toi8(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32 u = *(wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));

	if(u<=127)
	{
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_cu32tou8(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32 u = *(wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));

	if(u<=255)
	{
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_cu32toi16(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32 u = *(wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));

	if(u<=32767)
	{
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_cu32tou16(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32 u = *(wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));

	if(u<=65535)
	{
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_ci64toi8(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_i64 u = *(wt_i64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));

	if((-128)<=u && u<=127)
	{
		ctx->dsp += sizeof(wt_u64) - sizeof(wt_u32);
		*(wt_u32*)ctx->dsp = ((wt_u32)((wt_u64)u)) & 0xFFu;

		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_ci64tou8(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_i64 u = *(wt_i64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));

	if(0<=u && u<=255)
	{
		ctx->dsp += sizeof(wt_u64) - sizeof(wt_u32);
		*(wt_u32*)ctx->dsp = ((wt_u32)((wt_u64)u));

		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_ci64toi16(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_i64 u = *(wt_i64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));

	if((-32768)<=u && u<=32767)
	{
		ctx->dsp += sizeof(wt_u64) - sizeof(wt_u32);
		*(wt_u32*)ctx->dsp = ((wt_u32)((wt_u64)u)) & 0xFFFFu;

		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_ci64tou16(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_i64 u = *(wt_i64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));

	if(0<=u && u<=65535)
	{
		ctx->dsp += sizeof(wt_u64) - sizeof(wt_u32);
		*(wt_u32*)ctx->dsp = ((wt_u32)((wt_u64)u));

		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_cu64toi8(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u64 u = *(wt_u64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));

	if(u<=127)
	{
		ctx->dsp += sizeof(wt_u64) - sizeof(wt_u32);
		*(wt_u32*)ctx->dsp = ((wt_u32)((wt_u64)u));

		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_cu64tou8(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u64 u = *(wt_u64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));

	if(u<=255)
	{
		ctx->dsp += sizeof(wt_u64) - sizeof(wt_u32);
		*(wt_u32*)ctx->dsp = ((wt_u32)((wt_u64)u));

		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_cu64toi16(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u64 u = *(wt_u64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));

	if(u<=32767)
	{
		ctx->dsp += sizeof(wt_u64) - sizeof(wt_u32);
		*(wt_u32*)ctx->dsp = ((wt_u32)((wt_u64)u));

		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_cu64tou16(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u64 u = *(wt_u64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));

	if(u<=65535)
	{
		ctx->dsp += sizeof(wt_u64) - sizeof(wt_u32);
		*(wt_u32*)ctx->dsp = ((wt_u32)((wt_u64)u));

		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_ci8tou8(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32 u = *(wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));

	if(u<=127)
	{
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_cu8toi8(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	return wtr_mp_ci8tou8(prof, prog, ctx);
}

enum wtr_runresult wtr_mp_ci16toi8(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32 u = *(wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));

	if(u<=127)
	{
		return WTRR_RUN;
	}
	else
	if(0xFF80<u && u<=0xFFFF)
	{
		*(wt_u32*)ctx->dsp &= 0xFF;
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_ci16tou8(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32 u = *(wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));

	if(u<=127)
	{
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_ci16tou16(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32 u = *(wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));

	if(u<=0x7FFF)
	{
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_cu16toi16(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	return wtr_mp_ci16tou16(prof, prog, ctx);
}

enum wtr_runresult wtr_mp_cu16toi8(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32 u = *(wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));

	if(u<=127)
	{
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_cu16tou8(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32 u = *(wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));

	if(u<=255)
	{
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_version(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	((void)prof);
	((void)prog);

	CHECK_PUSH(ctx, sizeof(wt_u32));
	WTR_CTX_DS_PUSH(ctx, wt_u32, WTR_VERSION);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_compversion(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	((void)prof);
	((void)prog);

	CHECK_PUSH(ctx, sizeof(wt_u32));
	WTR_CTX_DS_PUSH(ctx, wt_u32, WTR_BINCOMPSINCEVER);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_coreversion(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	return wtr_mp_version(prof, prog, ctx);
}

enum wtr_runresult wtr_mp_nullhandle(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	((void)prof);
	((void)prog);

	CHECK_PUSH(ctx, sizeof(wt_handle));
	WTR_CTX_DS_PUSH(ctx, wt_handle, 0);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_handleequal(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_handle lhs;
	wt_handle rhs;
	wt_bool   res = WT_TRUE;
	CHECK_POP(ctx, sizeof(wt_handle)*2);
	lhs = *(wt_handle*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_handle);
	rhs = *(wt_handle*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_handle);

	((void)prof);
	((void)prog);

	if(lhs!=rhs)
	{
		res = WT_FALSE;
	}

	WTR_CTX_DS_PUSH(ctx, wt_bool, res);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_fileaccess(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_vmptr        vma;
	wt_i32          mode;
	struct wtr_str* str;

	CHECK_POP(ctx, sizeof(wt_vmptr)+sizeof(wt_i32));

	((void)prof);

	mode = *(wt_i32*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_i32);
	vma = *(wt_vmptr*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_vmptr);

	str = ((struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, vma));
	ASSERT_OBJECT_EXISTENCE(str);

	wtr_iocall_access(ctx, str->data, mode);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_filestat(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_vmptr        vma;
	struct wtr_str* str;

	CHECK_POP(ctx, sizeof(wt_vmptr));

	((void)prof);

	vma = *(wt_vmptr*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_vmptr);

	str = ((struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, vma));
	ASSERT_OBJECT_EXISTENCE(str);

	wtr_iocall_stat(ctx, str->data);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_filestatf(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_file f;

	CHECK_POP(ctx, sizeof(wt_vmptr));

	((void)prof);

	f = (wt_file)(wt_uintptr)*(wt_handle*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_handle);

	wtr_iocall_fstat(ctx, f);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_fileopen(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_vmptr        vma;
	wt_char         ch;
	struct wtr_str* str;

	CHECK_POP(ctx, sizeof(wt_vmptr)+sizeof(wt_char));

	((void)prof);

	ch = *(wt_char*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_char);
	vma = *(wt_vmptr*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_vmptr);

	str = ((struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, vma));
	ASSERT_OBJECT_EXISTENCE(str);

	wtr_iocall_fopen(ctx, str->data, (char)ch);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_fileclose(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_file f;

	CHECK_POP(ctx, sizeof(wt_handle));

	((void)prof);

	f = (wt_file)(wt_uintptr)*(wt_handle*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_handle);

	wtr_iocall_fclose(ctx, f);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_fileeof(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_file f;

	CHECK_POP(ctx, sizeof(wt_handle));

	((void)prof);

	f = (wt_file)(wt_uintptr)*(wt_handle*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_handle);

	wtr_iocall_feof(ctx, f);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_fileread(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_file  f;
	wt_vmptr addr;
	wt_u32   size;

	CHECK_POP(ctx, sizeof(wt_handle)+sizeof(wt_vmptr)+sizeof(wt_u32));

	((void)prof);

	size = *(wt_u32*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_u32);
	addr = *(wt_vmptr*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_vmptr);
	f = (wt_file)(wt_uintptr)*(wt_handle*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_handle);

	wtr_iocall_fread(ctx, f, WTR_PTR_FROM_VMADDR(prog, addr), size, 1);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_filereadr(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_file  f;
	wt_vmptr addr;
	wt_u32   size;

	CHECK_POP(ctx, sizeof(wt_handle)+sizeof(wt_vmptr)+sizeof(wt_u32));

	((void)prof);

	size = *(wt_u32*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_u32);
	addr = *(wt_vmptr*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_vmptr);
	f = (wt_file)(wt_uintptr)*(wt_handle*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_handle);

	wtr_iocall_fread(ctx, f, WTR_PTR_FROM_VMADDR(prog, addr), size, 0);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_filereadstr(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_file                        f;
	wt_char                        sep;

	CHECK_POP(ctx, sizeof(wt_handle));

	((void)prof);

	sep = *(wt_char*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_char);
	f = (wt_file)(wt_uintptr)*(wt_handle*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_handle);

	wtr_iocall_freadln(ctx, f, ctx->sbuf, ctx->sbuf_size, (char)sep);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_filewrite(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_file  f;
	wt_vmptr addr;
	wt_u32   size;

	CHECK_POP(ctx, sizeof(wt_handle)+sizeof(wt_vmptr)+sizeof(wt_u32));

	((void)prof);

	size = *(wt_u32*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_u32);
	addr = *(wt_vmptr*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_vmptr);
	f = (wt_file)(wt_uintptr)*(wt_handle*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_handle);

	wtr_iocall_fwrite(ctx, f, WTR_PTR_FROM_VMADDR(prog, addr), size, 1);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_filewritestr(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_file         f;
	wt_char         sep;
	struct wtr_str* s;

	CHECK_POP(ctx, sizeof(wt_vmptr)+sizeof(wt_char)+sizeof(wt_handle));

	((void)prof);

	sep = *(wt_char*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_char);
	s = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
	WTR_CTX_DS_POP(ctx, wt_vmptr);
	f = (wt_file)(wt_uintptr)*(wt_handle*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_handle);

	wtr_iocall_fwriteln(ctx, f, s->data, s->length, (char)sep);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_diropen(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_vmptr vma;

	CHECK_POP(ctx, sizeof(wt_vmptr));

	((void)prof);

	vma = *(wt_vmptr*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_vmptr);

	wtr_iocall_opendir(ctx,
		((struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, vma))->data);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_dirclose(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_dir d;

	CHECK_POP(ctx, sizeof(wt_handle));

	((void)prof);

	d = (wt_dir)(wt_uintptr)*(wt_handle*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_handle);

	wtr_iocall_closedir(ctx, d);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_dirread(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_dir d;

	CHECK_POP(ctx, sizeof(wt_handle));

	((void)prof);

	d = (wt_dir)(wt_uintptr)*(wt_handle*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_handle);

	wtr_iocall_readdir(ctx, d);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_dirgetname(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_dir d;

	CHECK_POP(ctx, sizeof(wt_handle));

	((void)prof);

	d = (wt_dir)(wt_uintptr)*(wt_handle*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_handle);

	wtr_iocall_dirgetname(ctx, d, ctx->sbuf, ctx->sbuf_size);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_dirgettype(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_dir d;

	CHECK_POP(ctx, sizeof(wt_handle));

	((void)prof);

	d = (wt_dir)(wt_uintptr)*(wt_handle*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_handle);

	wtr_iocall_dirgettype(ctx, d);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_rand(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32 num = _wtr_xorshift(prof);

	((void)prof);
	((void)prog);

	CHECK_PUSH(ctx, sizeof(wt_u32));
	WTR_CTX_DS_PUSH(ctx, wt_u32, num);

	return WTRR_RUN;
}

#if WTRUNTIME_ENABLE_FLOAT32

enum wtr_runresult wtr_mp_randf(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32 num = _wtr_xorshift(prof);

	((void)prof);
	((void)prog);

	CHECK_PUSH(ctx, sizeof(wt_f32));
	WTR_CTX_DS_PUSH(ctx, wt_f32, ((float)(num&0x00FFFFFF)) / ((float)0x00FFFFFF));

	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT32 */

enum wtr_runresult wtr_mp_srand(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32 seed;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(seed));
	seed = *(wt_u32*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_u32);

	prof->xs_x = (seed);
	prof->xs_y = (seed >> 13);
	prof->xs_z = (seed << 7);
	prof->xs_w = (seed ^ 0xa7a7a7a7);

	_wtr_xorshift(prof);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_islittleendian(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	((void)prof);
	((void)prog);

	CHECK_PUSH(ctx, sizeof(wt_bool));
	WTR_CTX_DS_PUSH(ctx, wt_bool, wtr_byteorder()==WTBO_LITTLE);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_isbigendian(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	((void)prof);
	((void)prog);

	CHECK_PUSH(ctx, sizeof(wt_bool));
	WTR_CTX_DS_PUSH(ctx, wt_bool, wtr_byteorder()==WTBO_BIG);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_swap16(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));
	*(wt_u32*)ctx->dsp = wtr_swap16((wt_u16)*(wt_u32*)ctx->dsp);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_swap32(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));
	*(wt_u32*)ctx->dsp = wtr_swap32(*(wt_u32*)ctx->dsp);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_swap64(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));
	*(wt_u64*)ctx->dsp = wtr_swap64(*(wt_u64*)ctx->dsp);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_lswap16(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));
	*(wt_u32*)ctx->dsp = wtr_botoh16((wt_u16)*(wt_u32*)ctx->dsp, WTBO_LITTLE);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_lswap32(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));
	*(wt_u32*)ctx->dsp = wtr_botoh32(*(wt_u32*)ctx->dsp, WTBO_LITTLE);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_lswap64(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));
	*(wt_u64*)ctx->dsp = wtr_botoh64(*(wt_u64*)ctx->dsp, WTBO_LITTLE);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_bswap16(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));
	*(wt_u32*)ctx->dsp = wtr_botoh16((wt_u16)*(wt_u32*)ctx->dsp, WTBO_BIG);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_bswap32(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u32));
	*(wt_u32*)ctx->dsp = wtr_botoh32(*(wt_u32*)ctx->dsp, WTBO_BIG);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_bswap64(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_u64));
	*(wt_u64*)ctx->dsp = wtr_botoh64(*(wt_u64*)ctx->dsp, WTBO_BIG);

	return WTRR_RUN;
}

static enum wtr_runresult wtr_mp__chrclassify_impl(
	struct wtr_profile* const prof, struct wtr_program* const prog,
	struct wtr_context* const ctx, int (*chrclassify)(int))
{
	const wt_char ch = *(wt_char*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(wt_char));
	CHECK_PUSH(ctx, sizeof(wt_bool));

	ctx->dsp -= sizeof(wt_bool) - sizeof(wt_char);
	*(wt_bool*)ctx->dsp = (wt_bool)chrclassify((char)ch);

	return WTRR_RUN;
}

#define _DEF_CHAR_CLASSIFICATION_FUNC(_name_) \
enum wtr_runresult wtr_mp_##_name_(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx) \
{ \
	return wtr_mp__chrclassify_impl(prof, prog, ctx, _name_); \
}

_DEF_CHAR_CLASSIFICATION_FUNC(iscntrl)
_DEF_CHAR_CLASSIFICATION_FUNC(isprint)
_DEF_CHAR_CLASSIFICATION_FUNC(isspace)
_DEF_CHAR_CLASSIFICATION_FUNC(isblank)
_DEF_CHAR_CLASSIFICATION_FUNC(isgraph)
_DEF_CHAR_CLASSIFICATION_FUNC(ispunct)
_DEF_CHAR_CLASSIFICATION_FUNC(isalnum)
_DEF_CHAR_CLASSIFICATION_FUNC(isalpha)
_DEF_CHAR_CLASSIFICATION_FUNC(isupper)
_DEF_CHAR_CLASSIFICATION_FUNC(islower)
_DEF_CHAR_CLASSIFICATION_FUNC(isdigit)
_DEF_CHAR_CLASSIFICATION_FUNC(isxdigit)

static wt_u32 wtr_clz32(wt_u32 x)
{
	wt_u32 n = 1;

	if(0==x) { return 32; }
	if(0==(x>>16)) { n += 16; x <<= 16; }
	if(0==(x>>24)) { n +=  8; x <<=  8; }
	if(0==(x>>28)) { n +=  4; x <<=  4; }
	if(0==(x>>30)) { n +=  2; x <<=  2; }
	n -= (x>>31);

	return n;
}

static wt_u32 wtr_ctz32(wt_u32 x)
{
	wt_u32 n = 1;

	if(0==x) { return 32; }
	if(0==(x & (wt_u32)(0x0000FFFFu))) { n += 16; x >>= 16;}
	if(0==(x & (wt_u32)(0x000000FFu))) { n +=  8; x >>=  8;}
	if(0==(x & (wt_u32)(0x0000000Fu))) { n +=  4; x >>=  4;}
	if(0==(x & (wt_u32)(0x00000003u))) { n +=  2; x >>=  2;}

	return n - (x & 1);
}

static wt_u64 wtr_clz64(wt_u64 x)
{
	wt_u64 n = 1;

	if(0==x) { return 64; }
	if(0==(x>>32)) { n += 32; x <<= 32; }
	if(0==(x>>48)) { n += 16; x <<= 16; }
	if(0==(x>>56)) { n +=  8; x <<=  8; }
	if(0==(x>>60)) { n +=  4; x <<=  4; }
	if(0==(x>>62)) { n +=  2; x <<=  2; }
	n -= (x>>63);

	return n;
}

static wt_u64 wtr_ctz64(wt_u64 x)
{
	wt_u64 n = 1;

	if(0==x) { return 64; }
	if(0==(x & (wt_u64)(0xFFFFFFFFul))) { n += 32; x >>= 32;}
	if(0==(x & (wt_u64)(0x0000FFFFul))) { n += 16; x >>= 16;}
	if(0==(x & (wt_u64)(0x000000FFul))) { n +=  8; x >>=  8;}
	if(0==(x & (wt_u64)(0x0000000Ful))) { n +=  4; x >>=  4;}
	if(0==(x & (wt_u64)(0x00000003ul))) { n +=  2; x >>=  2;}

	return n - (wt_u64)(x & 1);
}

enum wtr_runresult wtr_mp_clz8(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32* const top = (wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = wtr_clz32((*top) & 0xFF) - 24;
	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_clz16(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32* const top = (wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = wtr_clz32((*top) & 0xFFFF) - 16;
	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_clz32(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32* const top = (wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = wtr_clz32(*top);
	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_clz64(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u64* const top = (wt_u64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = wtr_clz64(*top);
	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_ctz8(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32* const top = (wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = wtr_ctz32((*top) | 0x100);
	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_ctz16(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32* const top = (wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = wtr_ctz32((*top) | 0x10000);
	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_ctz32(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u32* const top = (wt_u32*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = wtr_ctz32(*top);
	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_ctz64(struct wtr_profile* const prof, struct wtr_program* const prog, struct wtr_context* const ctx)
{
	wt_u64* const top = (wt_u64*)ctx->dsp;

	((void)prof);
	((void)prog);

	CHECK_POP(ctx, sizeof(*top));
	*top = wtr_ctz64(*top);
	return WTRR_RUN;
}
